C     Solves Krusell and Smith (1998) economy as in Young (2005)
C     Uses piecewise-histogram approximation to parameterize cross-sectional distribution.
C     Uses hybrid cubic spline-polynomial interpolation scheme to parameterize value function.
C     To run:  compile with links to helph36.f, matrix4.f, minim.f, random2.f, and statfunc.f
C     Required input files:  xkpts150.in, saving.f1
C     Hot start input files:  jedc_final_1.out3, jedc_final_1.out
C     Written:  ERY, June 2002
C     Last edited:  ERY, April 2005

      implicit real*8 (a-h,o-z)
      parameter (nk2pts=5000,nkpts=150,nread=1,nsample=1000,
     &           nmupts=4,ncapt=10000,nm2pts=30,durug=1.5D+00,
     &           unempg=0.04D+00,durgd=8.0D+00,unempb=0.1D+00,
     &           durbd=8.0D+00,zgood=1.01D+00,zbad=0.99D+00,
     &           durub=2.5D+00,alpha=0.36D+00,nread3=1,ndod=0,
     &           hfix=1.0D+00/0.9D+00,toler=1.0D-07,
     &           theta=0.25D+00,mxiter=100)
      common/cfs/f1(nk2pts,2)
      common/cv2/y2(nkpts,nmupts,4),optk12(nkpts,nmupts,2),
     &           optk02(nkpts,nmupts,2)
      common/cprice/r(nmupts,2),w(nmupts,2)
	common/cgrid/xkpts(nkpts),xmupts(nmupts)
	common/ccoef/a(2),b(2),anew(2),bnew(2)
	common/cprob/pi(4,4)
	common/nshock/nz(ncapt+1),nind(ncapt)
	common/cmgrid/xm2pts(nm2pts)
	common/ckgrid/xk2pts(nk2pts)
	common/cdecg/optkg1(nk2pts,nm2pts,2),optkg0(nk2pts,nm2pts,2)
	allocatable :: uniz(:),hansens(:),dseeds(:),unie(:)
 
      one = 1.0D+00  

      pgg00 = (durug-one)/durug
	pbb00 = (durub-one)/durub
	pbg00 = 1.25D+00*pbb00
	pgb00 = 0.75D+00*pgg00
	pgg01 = (unempg-unempg*pgg00)/(one-unempg)
	pbb01 = (unempb-unempb*pbb00)/(one-unempb)
	pbg01 = (unempb-unempg*pbg00)/(one-unempg)
	pgb01 = (unempg-unempb*pgb00)/(one-unempb)
	pgg = (durgd-one)/durgd
	pgb = one-(durbd-one)/durbd

	pgg10 = one-(durug-one)/durug
	pbb10 = one-(durub-one)/durub
	pbg10 = one-1.25D+00*pbb00
	pgb10 = one-0.75D+00*pgg00
	pgg11 = one-(unempg-unempg*pgg00)/(one-unempg)
	pbb11 = one-(unempb-unempb*pbb00)/(one-unempb)
	pbg11 = one-(unempb-unempg*pbg00)/(one-unempg)
	pgb11 = one-(unempg-unempb*pgb00)/(one-unempb)
	pbg = one-(durgd-one)/durgd
	pbb = (durbd-one)/durbd

	pi(1,1) = pgg*pgg11
	pi(1,2) = pgg*pgg01
	pi(1,3) = pbg*pbg11
	pi(1,4) = pbg*pbg01

	pi(2,1) = pgg*pgg10
	pi(2,2) = pgg*pgg00
	pi(2,3) = pbg*pbg10
	pi(2,4) = pbg*pbg00

	pi(3,1) = pgb*pgb11
	pi(3,2) = pgb*pgb01
	pi(3,3) = pbb*pbb11
	pi(3,4) = pbb*pbb01

	pi(4,1) = pgb*pgb10
	pi(4,2) = pgb*pgb00
	pi(4,3) = pbb*pbb10
	pi(4,4) = pbb*pbb00

      write(6,592) ((pi(i,j),j=1,4),i=1,4)
 592  format(4f18.10)

      allocate (uniz(ncapt+1))

      dseed = 1894.0D+00
	call ggubs(dseed,ncapt+1,uniz)
	nz(1) = 1
	do 52 i = 1,ncapt
	   if (nz(i) .eq. 1) then
	      if (uniz(i) .le. pgg) then
	         nz(i+1) = 1
	      else
	         nz(i+1) = 2
	      endif
	   else
	      if (uniz(i) .le. pgb) then
	         nz(i+1) = 1
	      else
	         nz(i+1) = 2
	      endif
	   endif
 52   continue

      deallocate (uniz)
      allocate (unie(ncapt))

	nind(1) = 1
	dseed = 4682322.0D+00
	call ggubs(dseed,ncapt,unie)
	do 54 i = 1,ncapt-1
	   if (nz(i) .eq. 1 .and. nz(i+1) .eq. 1) then
	      if (nind(i) .eq. 1) then
	         if (unie(i) .le. pgg11) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      else
	         if (unie(i) .le. pgg10) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      endif
	   elseif (nz(i) .eq. 1 .and. nz(i+1) .eq. 2) then
	      if (nind(i) .eq. 1) then
	         if (unie(i) .le. pbg11) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      else
	         if (unie(i) .le. pbg10) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      endif
	   elseif (nz(i) .eq. 2 .and. nz(i+1) .eq. 1) then
	      if (nind(i) .eq. 1) then
	         if (unie(i) .le. pgb11) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      else
	         if (unie(i) .le. pgb10) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      endif
	   else
	      if (nind(i) .eq. 1) then
	         if (unie(i) .le. pbb11) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      else
	         if (unie(i) .le. pbb10) then
	            nind(i+1) = 1
	         else
	            nind(i+1) = 0
	         endif
	      endif
	   endif
 54   continue
      deallocate (unie)

      open(2,file='xkpts150.in',form='formatted')
      do 999 i = 1,nkpts
         read(2,570) xkpts(i)
 570     format(f18.10)
 999  continue
      close(2)
	xkpts = xkpts*3.0D+00
      xk2pts(1) = xkpts(1)
      xkinc = (1200.0D+00-xkpts(1))/dble(real(nk2pts-1))
      do 393 i = 2,nk2pts
         xk2pts(i) = xk2pts(i-1)+xkinc
 393  continue

      xmupts(1) = 36.0D+00
	xmupts(2) = 36.0D+00+(43.0D+00-36.0D+00)/3.0D+00
	xmupts(3) = 36.0D+00+2.0D+00*(43.0D+00-36.0D+00)/3.0D+00
	xmupts(4) = 43.0D+00
	xminc = (xmupts(4)-xmupts(1))/dble(real(nm2pts-1))
	xm2pts(1) = xmupts(1)
	do 58 i = 2,nm2pts
	   xm2pts(i) = xm2pts(i-1)+xminc
 58   continue
      do 38 i = 1,nmupts
         r(i,1) = alpha*zgood*xmupts(i)**(alpha-one)*
     &            ((one-unempg)*hfix)**(one-alpha)
	   w(i,1) = (one-alpha)*zgood*xmupts(i)**alpha*
     &            ((one-unempg)*hfix)**(-alpha)
	   r(i,2) = alpha*zbad*xmupts(i)**(alpha-one)*
     &            ((one-unempb)*hfix)**(one-alpha)
	   w(i,2) = (one-alpha)*zbad*xmupts(i)**alpha*
     &            ((one-unempb)*hfix)**(-alpha)
	   write(6,787) r(i,1),w(i,1),r(i,2),w(i,2)
 787     format(4f18.10)
 38   continue

      if (nread .eq. 1) then
         open(2,file='jedc_final_1.out',form='formatted')
         do 98 i = 1,nkpts
	      do 97 j = 1,nmupts
               read(2,700) dum,dum,y2(i,j,1),y2(i,j,2),
     &                     y2(i,j,3),y2(i,j,4),
     &                     optk12(i,j,1),optk12(i,j,2),
     &                     optk02(i,j,1),optk02(i,j,2)
 700           format(10f18.10)
 97         continue
 98      continue
         close(2)
      else
	   open(2,file='xkpts150.in',form='formatted')
         do 57 i = 1,nkpts
	      read(2,390) xkpts(i)
 390        format(f18.10)
 57      continue
         close(2)
	   do 56 i = 1,nkpts
            do 96 j = 1,nmupts
               y2(i,j,1) = dlog(xkpts(i)+
     &                     (one-alpha)/alpha*
     &                     xmupts(j)/(one-unempg))
	         y2(i,j,2) = dlog(xkpts(i)+
     &                     whome/r(j,1))
     	         y2(i,j,3) = dlog(xkpts(i)+
     &                     (one-alpha)/alpha*
     &                     xmupts(j)/(one-unempb))
	         y2(i,j,4) = dlog(xkpts(i)+
     &                     whome/r(j,2))
               optk12(i,j,1) = xkpts(i)
	         optk12(i,j,2) = xkpts(i)
               optk02(i,j,1) = xkpts(i)
	         optk02(i,j,2) = xkpts(i)
 96         continue
 56      continue
      endif

	if (nread3 .eq. 1) then
	   open(2,file='jedc_final_1.out3',form='formatted')
	   read(2,303) (a(i),i=1,2)
	   read(2,303) (b(i),i=1,2)
 303     format(2f18.10)
         close(2)
	else     
         a(1) = dlog(39.3D+00)*0.04D+00
	   a(2) = 0.96D+00
	   b = a
	endif

c     Computing equilibrium forecasting functions

      iter = 1
 1    call fvec(a,b,anew,bnew)
	write(6,303) (a(i),i=1,2)
	write(6,303) (anew(i),i=1,2)
	write(6,303) (b(i),i=1,2)
	write(6,303) (bnew(i),i=1,2)
	
	iter = iter+1
      diffmx = 0.0D+00
	do 74 i = 1,2
	   adiff = dabs(anew(i)-a(i))
	   if (adiff .gt. diffmx) diffmx = adiff
	   bdiff = dabs(bnew(i)-b(i))
	   if (bdiff .gt. diffmx) diffmx = bdiff
 74   continue
      if (diffmx .lt. toler .or. iter .gt. mxiter) goto 2

	do 63 i = 1,2
	   a(i) = (one-theta)*a(i)+theta*anew(i)
	   b(i) = (one-theta)*b(i)+theta*bnew(i)
 63   continue
      open(9,file='jedc_final_1.tmp',form='formatted')
      write(9,303) (a(i),i=1,2)
	write(9,303) (b(i),i=1,2)
      close(9)
      goto 1

 2    open(9,file='jedc_final_1.out3',form='formatted')
      write(9,303) (a(i),i=1,2)
	write(9,303) (b(i),i=1,2)
      close(9)

c     Distribution of den Haan-Marcet statistics

      if (ndod .eq. 1) then

         allocate (dseeds(nsample),hansens(nsample))

         eseed = 34958.0D+00
	   call ggubs(eseed,nsample,dseeds)

         hansens = 0.0D+00
	   do 39 j = 1,nsample
	      dseed = dseeds(j)*1000.0D+00
	      call ggubs(dseed,ncapt+1,uniz)
	      nz(1) = 1
	      do 53 i = 1,ncapt
	         if (nz(i) .eq. 1) then
	            if (uniz(i) .le. pgg) then
	               nz(i+1) = 1
	            else
	               nz(i+1) = 2
	            endif
	         else
	            if (uniz(i) .le. pgb) then
	               nz(i+1) = 1
	            else
	               nz(i+1) = 2
	            endif
	         endif
 53         continue
            call calcf(optkg1,optkg0,xk2pts,xm2pts,a,b,anew,bnew,
     &                 xkmean,sdmean,skmean,s4mean,s5mean,rsqa,rsqb,
     &                 sderra,sderrb,
     &                 gmean,p1mean,p5mean,p10mean,p20mean,
     &                 p30mean,p90mean,p95mean,hansen,erfmx1,erfmx2)
            hansens(j) = hansen
	      if (multpl(j,10) .eq. 1) print *,j
 39      continue

         open(2,file='jedc_final_1.hans',form='formatted')
	   do 88 i = 1,nsample
	      write(2,366) hansens(i)
 366        format(f18.10)
 88      continue
         close(2)

         deallocate (hansens,dseeds)

      endif

c     JEDC simulation experiments

      open(1,file='agg_switch.txt',form='formatted')
	do 256 i = 1,ncapt
	   read(1,381) nnz
 381     format(i1)
         if (nnz .eq. 1) then
	      nz(i) = 2
	   else
	      nz(i) = 1
	   endif
 256  continue
      close(1)
	
	open(1,file='ind_switch.txt',form='formatted')
	do 255 i = 1,ncapt
	   read(1,381) nnind
         if (nnind .eq. 1) then
	      nind(i) = 2
	   else
	      nind(i) = 1
	   endif
 255  continue
      close(1)
      call calcf3(optkg1,optkg0,xk2pts,xm2pts,a,b,
     &            xkmean,sdmean,skmean,s4mean,s5mean,rsqa,rsqb,
     &            sderra,sderrb,
     &            gmean,p1mean,p5mean,p10mean,p20mean,
     &            p30mean,p90mean,p95mean,hansen,erfmx1,erfmx2)

      nz(1:100) = 2
	nz(101:ncapt) = 1
      call calcf2(optkg1,optkg0,xk2pts,xm2pts,a,b,
     &            xkmean,sdmean,skmean,s4mean,s5mean,rsqa,rsqb,
     &            sderra,sderrb,
     &            gmean,p1mean,p5mean,p10mean,p20mean,
     &            p30mean,p90mean,p95mean,hansen,erfmx1,erfmx2)

	end

      subroutine fvec(a,b,anew,bnew)
      implicit real*8 (a-h,o-z)
      parameter (nkpts=150,alpha=0.36D+00,delta=0.025D+00,
     &           hfix=1.0D+00/0.9D+00,xkbor=0.0D+00,beta=0.99D+00,
     &           mxiter=2000,nk2pts=5000,nm2pts=30,gamma=1.0D+00,
     &           mxpol=200,nmupts=4,unempb=0.1D+00,xmu=0.15D+00,
     &	       unempg=0.04D+00,zgood=1.01D+00,zbad=0.99D+00)
      common/cfs/f1(nk2pts,2)
      common/cv2/y2(nkpts,nmupts,4),optk12(nkpts,nmupts,2),
     &           optk02(nkpts,nmupts,2)
      common/cprice/r(nmupts,2),w(nmupts,2)
	common/cgrid/xkpts(nkpts),xmupts(nmupts)
	common/cvs/v11(nkpts,nmupts,2),v11dp(nkpts,nmupts,2),
     &           v10(nkpts,nmupts,2),v10dp(nkpts,nmupts,2),
     &           v01(nkpts,nmupts,2),v01dp(nkpts,nmupts,2),
     &           v00(nkpts,nmupts,2),v00dp(nkpts,nmupts,2)
	common/cws/w11(nkpts,2),w11dp(nkpts,2),
     &           w10(nkpts,2),w10dp(nkpts,2),
     &           w01(nkpts,2),w01dp(nkpts,2),
     &           w00(nkpts,2),w00dp(nkpts,2)
      common/cprob/pi(4,4)
	common/cmgrid/xm2pts(nm2pts)
	common/ckgrid/xk2pts(nk2pts)
	common/cvgrid/vgrid(nk2pts,nm2pts,4)
	common/cdecg/optkg1(nk2pts,nm2pts,2),optkg0(nk2pts,nm2pts,2)
	allocatable :: y(:,:,:),optk1(:,:,:),optk0(:,:,:)
      dimension a(2),b(2),anew(2),bnew(2),r0(2),w0(2)

      one = 1.0D+00
      xkcut1 = 0.5D+00
      xkcut2 = 0.5D+00

      xklow = xkbor
      vmxdif = 100.0D+00

      allocate (y(nkpts,nmupts,4),optk1(nkpts,nmupts,2),
     &          optk0(nkpts,nmupts,2))

c     Value Iteration

      do 30 kkk = 1,mxiter

         if (vmxdif .lt. 1.0D-00) then
	      ncalop = 0
	   else
	      ncalop = 1
	   endif
                 
         y = y2
	   optk1 = optk12
	   optk0 = optk02

         call interp(xkpts,xmupts,y,v11,v10,v01,v00,
     &               v11dp,v10dp,v01dp,v00dp,a,b)

         do 13 i = 1,nkpts
	      do 567 j = 1,nmupts
               xkcur = xkpts(i)
	         nmu = j

               neps = 1
	         nagg = 1
               tau = xmu*unempg/(hfix*(one-unempg))
			 xkhigh = (r(j,1)+one-delta)*xkcur+(one-tau)*w(j,1)*hfix
			 if (xkcur .le. xkcut1 .or. ncalop .eq. 1) then
                  call calopt(xkcur,neps,xkp1,fval1,niter,
     &                        1.0D-10,r,w,xkhigh,xklow,
     &                        grad1,xkpts,xmupts,nmu,nagg)
               else
                  call altopt(xkcur,neps,xkp1,fval1,niter,
     &                        1.0D-10,optk1(i,j,1),grad1,
     &                        xkpts,r,w,xmupts,nmu,nagg,ndone)
                  if (ndone .eq. 0) then
                     call calopt(xkcur,neps,xkp1,fval1,niter,
     &                           1.0D-10,r,w,xkhigh,xklow,
     &                           grad1,xkpts,xmupts,nmu,nagg)
	            endif
               endif
               optk12(i,j,1) = xkp1
               y2(i,j,1) = fval1

               neps = 0
	         nagg = 1
               xkhigh = (r(j,1)+one-delta)*xkcur+xmu*w(j,1)
			 if (xkcur .le. xkcut2 .or. ncalop .eq. 1) then
                  call calopt(xkcur,neps,xkp2,fval2,niter,
     &                        1.0D-10,r,w,xkhigh,xklow,
     &                        grad0,xkpts,xmupts,nmu,nagg)
               else
                  call altopt(xkcur,neps,xkp2,fval2,niter,
     &                        1.0D-10,optk0(i,j,1),grad0,
     &                        xkpts,r,w,xmupts,nmu,nagg,ndone)
	            if (ndone .eq. 1) then
                     call calopt(xkcur,neps,xkp2,fval2,niter,
     &                           1.0D-10,r,w,xkhigh,xklow,
     &                           grad0,xkpts,xmupts,nmu,nagg)
                  endif
               endif
               optk02(i,j,1) = xkp2
               y2(i,j,2) = fval2

               neps = 1
	         nagg = 2
               tau = xmu*unempb/(hfix*(one-unempb))
               xkhigh = (r(j,2)+one-delta)*xkcur+(one-tau)*w(j,2)*hfix
			 if (xkcur .le. xkcut1 .or. ncalop .eq. 1) then
                  call calopt(xkcur,neps,xkp3,fval3,niter,
     &                        1.0D-10,r,w,xkhigh,xklow,
     &                        grad1,xkpts,xmupts,nmu,nagg)
               else
                  call altopt(xkcur,neps,xkp3,fval3,niter,
     &                        1.0D-10,optk1(i,j,2),grad1,
     &                        xkpts,r,w,xmupts,nmu,nagg,ndone)
	            if (ndone .eq. 1) then
                     call calopt(xkcur,neps,xkp3,fval3,niter,
     &                           1.0D-10,r,w,xkhigh,xklow,
     &                           grad1,xkpts,xmupts,nmu,nagg)
                  endif
               endif
               optk12(i,j,2) = xkp3
               y2(i,j,3) = fval3

               neps = 0
	         nagg = 2
               xkhigh = (r(j,2)+one-delta)*xkcur+xmu*w(j,2)
               if (xkcur .le. xkcut2 .or. ncalop .eq. 1) then
                  call calopt(xkcur,neps,xkp4,fval4,niter,
     &                        1.0D-10,r,w,xkhigh,xklow,
     &                        grad0,xkpts,xmupts,nmu,nagg)
               else
                  call altopt(xkcur,neps,xkp4,fval4,niter,
     &                        1.0D-10,optk0(i,j,2),grad0,
     &                        xkpts,r,w,xmupts,nmu,nagg,ndone)
	            if (ndone .eq. 0) then
                     call calopt(xkcur,neps,xkp4,fval4,niter,
     &                           1.0D-10,r,w,xkhigh,xklow,
     &                           grad0,xkpts,xmupts,nmu,nagg)
                  endif
               endif
               optk02(i,j,2) = xkp4
               y2(i,j,4) = fval4

 567        continue
 13      continue

c        Howard's improvement algorithm
	   
         if (vmxdif .lt. 1.0D+00) then
	      npol = 0
	   else
	      npol = 0
	   endif

	   if (npol .eq. 0) goto 2
         do 43 ii = 1,mxpol

	      call interp(xkpts,xmupts,y2,v11,v10,v01,v00,
     &                  v11dp,v10dp,v01dp,v00dp,a,b)

	      do 42 i = 1,nkpts
	         do 41 j = 1,nmupts
   	            xkcur = xkpts(i)

                  tau = xmu*unempg/(hfix*(one-unempg))
		        con1 = (r(j,1)+one-delta)*xkcur+(one-tau)*w(j,1)*hfix-
     &                   optk12(i,j,1)
	            call splint(xkpts,v11,v11dp,optk12(i,j,1),
     &                        yval1,yp,
     &                        ydp,khi,klo,0,nkpts,1,0,0,j,1)
	            call splint(xkpts,v10,v10dp,optk12(i,j,1),
     &                        yval2,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            call splint(xkpts,v01,v01dp,optk12(i,j,1),
     &                        yval3,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            call splint(xkpts,v00,v00dp,optk12(i,j,1),
     &                        yval4,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            if (dabs(gamma-one) .lt. 1.0D-04) then
	               utilc = dlog(con1)
	            else
	               utilc = one/(one-gamma)*(con1**(one-gamma)-one)
	            endif
	            y2(i,j,1) = utilc+beta*(pi(1,1)*yval1+
     &                        pi(1,2)*yval2+pi(1,3)*yval3+
     &                        pi(1,4)*yval4)

		        con2 = (r(j,1)+one-delta)*xkcur-optk02(i,j,1)+
     &                   xmu*w(j,1)
	            call splint(xkpts,v11,v11dp,optk02(i,j,1),
     &                        yval1,yp,
     &                        ydp,khi,klo,0,nkpts,1,0,0,j,1)
	            call splint(xkpts,v10,v10dp,optk02(i,j,1),
     &                        yval2,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            call splint(xkpts,v01,v01dp,optk02(i,j,1),
     &                        yval3,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            call splint(xkpts,v00,v00dp,optk02(i,j,1),
     &                        yval4,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,1)
	            if (dabs(gamma-one) .lt. 1.0D-04) then
	               utilc = dlog(con2)
	            else
	               utilc = one/(one-gamma)*(con2**(one-gamma)-one)
	            endif
	            y2(i,j,2) = utilc+beta*(pi(2,1)*yval1+
     &                        pi(2,2)*yval2+pi(2,3)*yval3+
     &                        pi(2,4)*yval4)

                  tau = xmu*unempb/(hfix*(one-unempb))
		        con3 = (r(j,2)+one-delta)*xkcur+(one-tau)*w(j,2)*hfix-
     &                   optk12(i,j,2)
	            call splint(xkpts,v11,v11dp,optk12(i,j,2),
     &                        yval1,yp,
     &                        ydp,khi,klo,0,nkpts,1,0,0,j,2)
	            call splint(xkpts,v10,v10dp,optk12(i,j,2),
     &                        yval2,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            call splint(xkpts,v01,v01dp,optk12(i,j,2),
     &                        yval3,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            call splint(xkpts,v00,v00dp,optk12(i,j,2),
     &                        yval4,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            if (dabs(gamma-one) .lt. 1.0D-04) then
	               utilc = dlog(con3)
	            else
	               utilc = one/(one-gamma)*(con3**(one-gamma)-one)
	            endif
	            y2(i,j,3) = utilc+beta*(pi(3,1)*yval1+
     &                        pi(3,2)*yval2+pi(3,3)*yval3+
     &                        pi(3,4)*yval4)

		        con4 = (r(j,2)+one-delta)*xkcur-optk02(i,j,2)+
     &                   xmu*w(j,2)
	            call splint(xkpts,v11,v11dp,optk02(i,j,2),
     &                        yval1,yp,
     &                        ydp,khi,klo,0,nkpts,1,0,0,j,2)
	            call splint(xkpts,v10,v10dp,optk02(i,j,2),
     &                        yval2,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            call splint(xkpts,v01,v01dp,optk02(i,j,2),
     &                        yval3,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            call splint(xkpts,v00,v00dp,optk02(i,j,2),
     &                        yval4,yp,
     &                        ydp,khi,klo,1,nkpts,1,0,0,j,2)
	            if (dabs(gamma-one) .lt. 1.0D-04) then
	               utilc = dlog(con4)
	            else
	               utilc = one/(one-gamma)*(con4**(one-gamma)-one)
	            endif
	            y2(i,j,4) = utilc+beta*(pi(4,1)*yval1+
     &                        pi(4,2)*yval2+pi(4,3)*yval3+
     &                        pi(4,4)*yval4)

 41            continue
 42         continue
 43      continue

 2       vmxdif = maxval(dabs(y2-y))
         dmxdif = maxval(dabs(optk12-optk1))
         emxdif = maxval(dabs(optk02-optk0))
         if (vmxdif .lt. 1.0D-06 .and. dmxdif .lt. 1.0D-08
     &       .and. emxdif .lt. 1.0D-08 .and. kkk .gt. 2) goto 8

         if (multpl(kkk,1) .eq. 1) then
            write(6,888) kkk,mxiter,vmxdif,dmxdif,emxdif
 888        format(2i5,3f15.10)
         endif

 30   continue

 8    write(6,888) kkk,mxiter,vmxdif,dmxdif,emxdif

      deallocate (y,optk1,optk0)

      open(3,file='jedc_final_1.out',form='formatted')
      do 45 i = 1,nkpts
	   do 44 j = 1,nmupts
            write(3,456) xkpts(i),xmupts(j),y2(i,j,1),
     &                   y2(i,j,2),y2(i,j,3),y2(i,j,4),
     &                   optk12(i,j,1),optk12(i,j,2),
     &                   optk02(i,j,1),optk02(i,j,2)
 456        format(10f18.10)
 44      continue
 45   continue
      close(3)

      open(3,file='jedc_final_1.out2',form='formatted')
      do 75 i = 1,nkpts
	   do 74 j = 1,nmupts
            write(3,457) xkpts(i),xmupts(j),y2(i,j,1),
     &                   y2(i,j,2),y2(i,j,3),y2(i,j,4)
 457        format(6f18.10)
 74      continue
 75   continue
      close(3)

c     Finer grid
      
      do 78 j = 1,nm2pts

         xmmu = xm2pts(j)
	   r0(1) = zgood*alpha*xm2pts(j)**(alpha-one)*
     &           ((one-unempg)*hfix)**(one-alpha)
	   w0(1) = zgood*(one-alpha)*xm2pts(j)**alpha*
     &           ((one-unempg)*hfix)**(-alpha)
	   r0(2) = zbad*alpha*xm2pts(j)**(alpha-one)*
     &           ((one-unempb)*hfix)**(one-alpha)
	   w0(2) = zbad*(one-alpha)*xm2pts(j)**alpha*
     &           ((one-unempb)*hfix)**(-alpha)

         call inter2(xkpts,xmupts,y2,w11,w10,w01,w00,
     &               w11dp,w10dp,w01dp,w00dp,a,b,xmmu)

         do 77 i = 1,nk2pts
         
            xkcur = xk2pts(i)

            neps = 1
	      nagg = 1
	      tau = xmu*unempg/(hfix*(one-unempg))
            xkhigh = (r0(1)+one-delta)*xkcur+(one-tau)*w0(1)*hfix
		  if (xkcur .le. xkcut1) then
               call calopt2(xkcur,neps,xkp1,fval1,niter,
     &                      1.0D-10,r0,w0,xkhigh,xklow,
     &                      grad1,xkpts,xmupts,xmu,nagg)
            else  
	         if (i .eq. 1) then
	            sk = xkcur
	         else
	            sk = optkg1(i-1,j,1)
	         endif
               call altopt2(xkcur,neps,xkp1,fval1,niter,
     &                      1.0D-10,sk,grad1,xkpts,r0,w0,xmupts,
     &                      xmu,nagg,ndone)
	         if (ndone .eq. 0) then
                  call calopt2(xkcur,neps,xkp1,fval1,niter,
     &                         1.0D-10,r0,w0,xkhigh,xklow,
     &                         grad1,xkpts,xmupts,xmu,nagg)
               endif 
            endif
            optkg1(i,j,1) = xkp1
            vgrid(i,j,1) = fval1

            neps = 0
	      nagg = 1
            xkhigh = (r0(1)+one-delta)*xkcur+xmu*w0(1)
            if (xkcur .le. xkcut2) then
               call calopt2(xkcur,neps,xkp2,fval2,niter,
     &                      1.0D-10,r0,w0,xkhigh,xklow,
     &                      grad0,xkpts,xmupts,xmu,nagg)
            else
	         if (i .eq. 1) then
	            sk = xkcur
	         else
	            sk = optkg0(i-1,j,1)
	         endif
               call altopt2(xkcur,neps,xkp2,fval2,niter,
     &                      1.0D-10,sk,grad0,xkpts,r0,w0,xmupts,
     &                      xmu,nagg,ndone)
	         if (ndone .eq. 0) then
                  call calopt2(xkcur,neps,xkp2,fval2,niter,
     &                         1.0D-10,r0,w0,xkhigh,xklow,
     &                         grad0,xkpts,xmupts,xmu,nagg)
               endif
            endif
            optkg0(i,j,1) = xkp2
            vgrid(i,j,2) = fval2

            neps = 1
	      nagg = 2
	      tau = xmu*unempb/(hfix*(one-unempb))
            xkhigh = (r0(2)+one-delta)*xkcur+(one-tau)*w0(2)*hfix
            if (xkcur .le. xkcut1) then
               call calopt2(xkcur,neps,xkp3,fval3,niter,
     &                      1.0D-10,r0,w0,xkhigh,xklow,
     &                      grad1,xkpts,xmupts,xmu,nagg)
            else  
	         if (i .eq. 1) then
	            sk = xkcur
	         else
	            sk = optkg1(i-1,j,2)
	         endif
               call altopt2(xkcur,neps,xkp3,fval3,niter,
     &                      1.0D-10,sk,grad1,xkpts,r0,w0,xmupts,
     &                      xmu,nagg,ndone)
               if (ndone .eq. 0) then
                  call calopt2(xkcur,neps,xkp3,fval3,niter,
     &                         1.0D-10,r0,w0,xkhigh,xklow,
     &                         grad1,xkpts,xmupts,xmu,nagg)
               endif
            endif
            optkg1(i,j,2) = xkp3
            vgrid(i,j,3) = fval3

            neps = 0
	      nagg = 2
            xkhigh = (r0(2)+one-delta)*xkcur+xmu*w0(2)
            if (xkcur .le. xkcut2) then
               call calopt2(xkcur,neps,xkp4,fval4,niter,
     &                      1.0D-10,r0,w0,xkhigh,xklow,
     &                      grad0,xkpts,xmupts,xmu,nagg)
            else
	         if (i .eq. 1) then
	            sk = xkcur
	         else
	            sk = optkg0(i-1,j,2)
	         endif
               call altopt2(xkcur,neps,xkp4,fval4,niter,
     &                      1.0D-10,sk,grad0,xkpts,r0,w0,xmupts,
     &                      xmu,nagg,ndone)
	         if (ndone .eq. 0) then
                  call calopt2(xkcur,neps,xkp4,fval4,niter,
     &                         1.0D-10,r0,w0,xkhigh,xklow,
     &                         grad0,xkpts,xmupts,xmu,nagg)
               endif
            endif
            optkg0(i,j,2) = xkp4
            vgrid(i,j,4) = fval4

 77      continue
 78   continue

      open(5,file='jedc_final_1.vgr',form='formatted')
      do 71 i = 1,nk2pts
	   do 70 j = 1,nm2pts
            write(5,603) xk2pts(i),xm2pts(j),
     &                   (vgrid(i,j,k),k=1,4)
 603        format(6f18.10)
 70      continue
 71   continue
      close(5)

      open(9,file='jedc_final_1.grd',form='formatted')
      do 99 i = 1,nk2pts
	   do 98 j = 1,nm2pts
            write(9,400) xk2pts(i),xm2pts(j),optkg1(i,j,1),
     &                   optkg1(i,j,2),optkg0(i,j,1),
     &                   optkg0(i,j,2)
 400        format(6f18.10)
 98      continue
 99   continue
      close(9)

c     Simulation

      write(6,7984)
 7984 format('computing simulation')

      call calcf(optkg1,optkg0,xk2pts,xm2pts,a,b,anew,bnew,
     &           xkmean,sdmean,skmean,s4mean,s5mean,rsqa,rsqb,sderra,
     &           sderrb,gmean,p1mean,p5mean,p10mean,p20mean,p30mean,
     &           p90mean,p95mean,hansen,erfmx1,erfmx2)

	write(6,799)
 799  format('moments of ergodic distribution')
      write(6,949) xkmean,sdmean,skmean,s4mean,s5mean
      write(6,949) p1mean,p5mean,p10mean,p20mean,p30mean
	write(6,949) p90mean,p95mean
	write(6,949) hansen,erfmx1/xkmean,erfmx2/xkmean,gmean
 744  format(f18.10)
 949  format(5f15.7)
      write(6,544)
 544  format('measures of fit')
      write(6,949) rsqa,rsqb,sderra,sderrb

      open(2,file='jedc_final_1.tmp2',form='formatted')
	write(2,949) rsqa,rsqb,sderra,sderrb
	close(2)

      return
      end

C-------------------------------------------------------------------

      subroutine splint(xa,ya,y2a,x,y,yp,ydp,khi,klo,kvals,n,
     &                  ndoy,ndoyp,ndoydp,nmu,nagg)
      implicit real*8 (a-h,o-z)
	parameter (nmupts=4,m=nmupts)
      dimension xa(n),ya(n,m,2),y2a(n,m,2)

      one = 1.0D+00
      if (kvals .eq. 0) then
         klo = 1
         khi = n
 1       if ((khi-klo) .gt. 1) then
            k = (khi+klo)/2
            if (xa(k) .gt. x) then
               khi = k
            else
               klo = k
            endif
            goto 1
         endif
      endif

      h = xa(khi)-xa(klo)
      a = (xa(khi)-x)/h
      b = (x-xa(klo))/h
      asq = a*a
      bsq = b*b

      if (ndoy .eq. 1)
     &   y = a*ya(klo,nmu,nagg)+b*ya(khi,nmu,nagg)+
     &       ((asq*a-a)*y2a(klo,nmu,nagg)+
     &       (bsq*b-b)*y2a(khi,nmu,nagg))*(h*h)/6.0D+00
      if (ndoyp .eq. 1)
     &   yp = (ya(khi,nmu,nagg)-ya(klo,nmu,nagg))/h-
     &        (3.0D+00*asq-one)/6.0D+00*h*y2a(klo,nmu,nagg)+
     &        (3.0D+00*bsq-one)/6.0D+00*h*y2a(khi,nmu,nagg)
      if (ndoydp .eq. 1)
     &   ydp = a*y2a(klo,nmu,nagg)+b*y2a(khi,nmu,nagg)
      
      return
      end

      subroutine splin2(xa,ya,y2a,x,y,yp,ydp,khi,klo,kvals,n,
     &                  ndoy,ndoyp,ndoydp,nagg)
      implicit real*8 (a-h,o-z)
	parameter (nmupts=4,m=nmupts)
      dimension xa(n),ya(n,2),y2a(n,2)

      one = 1.0D+00
      if (kvals .eq. 0) then
         klo = 1
         khi = n
 1       if ((khi-klo) .gt. 1) then
            k = (khi+klo)/2
            if (xa(k) .gt. x) then
               khi = k
            else
               klo = k
            endif
            goto 1
         endif
      endif

      h = xa(khi)-xa(klo)
      a = (xa(khi)-x)/h
      b = (x-xa(klo))/h
      asq = a*a
      bsq = b*b

      if (ndoy .eq. 1)
     &   y = a*ya(klo,nagg)+b*ya(khi,nagg)+
     &       ((asq*a-a)*y2a(klo,nagg)+
     &       (bsq*b-b)*y2a(khi,nagg))*(h*h)/6.0D+00
      if (ndoyp .eq. 1)
     &   yp = (ya(khi,nagg)-ya(klo,nagg))/h-
     &        (3.0D+00*asq-one)/6.0D+00*h*y2a(klo,nagg)+
     &        (3.0D+00*bsq-one)/6.0D+00*h*y2a(khi,nagg)
      if (ndoydp .eq. 1)
     &   ydp = a*y2a(klo,nagg)+b*y2a(khi,nagg)
      
      return
      end

      subroutine interp(x1,x2,y,y11,y10,y01,y00,y11dp,y10dp,
     &                  y01dp,y00dp,a0,b0)
      implicit real*8 (a-h,o-z)
      parameter (nkpts=150,n=nkpts,nmupts=4,m=nmupts)

      dimension x1(n),y(n,m,4),a(n),b(n),c(n),r(n),u(n),xx(n),
     &          yy(n),y11(n,m,2),y10(n,m,2),y11dp(n,m,2),
     &          y10dp(n,m,2),x2(m),y01(n,m,2),y00(n,m,2),
     &          y01dp(n,m,2),y00dp(n,m,2),a0(2),b0(2),
     &          xmup(m,2),xa(m),ya(m)

      do 72 i = 1,n
         xx(i) = x1(i)
 72   continue

      do 70 i = 1,m
	   xmup(i,1) = dexp(a0(1)+a0(2)*dlog(x2(i)))
	   xmup(i,2) = dexp(b0(1)+b0(2)*dlog(x2(i)))
 70   continue
      do 71 i = 1,m
	   xa(i) = x2(i)
 71   continue

      a(1) = 0.0D+00
      a(n) = -1.0D+00
      b(1) = 1.0D+00
      b(n) = 1.05D+00
      c(1) = -1.05D+00
      c(n) = 0.0D+00
      r(1) = 0.0D+00
      r(n) = 0.0D+00

      do 81 k = 1,4
	   do 80 ii = 1,m
	      do 79 jj = 1,2
               do 30 i = 1,n
	            do 29 j = 1,m
	               ya(j) = y(i,j,k)
 29               continue
                  call polint(xa,ya,m,xmup(ii,jj),yval,dy)
                  yy(i) = yval
                  if (k .eq. 1) y11(i,ii,jj) = yy(i)
                  if (k .eq. 2) y10(i,ii,jj) = yy(i)
	            if (k .eq. 3) y01(i,ii,jj) = yy(i)
	            if (k .eq. 4) y00(i,ii,jj) = yy(i)
 30            continue

               do 21 i = 2,n-1
                  a(i) = (xx(i)-xx(i-1))/6.0D+00
                  b(i) = (xx(i+1)-xx(i-1))/3.0D+00
                  c(i) = (xx(i+1)-xx(i))/6.0D+00
                  r(i) = (yy(i+1)-yy(i))/(xx(i+1)-xx(i))-
     &                   (yy(i)-yy(i-1))/(xx(i)-xx(i-1))
 21            continue

               call tridag(a,b,c,r,u,n)
               do 22 i = 1,n
                  if (k .eq. 1) y11dp(i,ii,jj) = u(i)
                  if (k .eq. 2) y10dp(i,ii,jj) = u(i)
	            if (k .eq. 3) y01dp(i,ii,jj) = u(i)
	            if (k .eq. 4) y00dp(i,ii,jj) = u(i)
 22            continue
 79         continue
 80      continue
 81   continue

      return
      end

      subroutine inter2(x1,x2,y,y11,y10,y01,y00,y11dp,y10dp,
     &                  y01dp,y00dp,a0,b0,xmu)
      implicit real*8 (a-h,o-z)
      parameter (nkpts=150,n=nkpts,nmupts=4,m=nmupts)

      dimension x1(n),y(n,m,4),a(n),b(n),c(n),r(n),u(n),xx(n),
     &          yy(n),y11(n,2),y10(n,2),y11dp(n,2),ya(m),
     &          y10dp(n,2),x2(m),y01(n,2),y00(n,2),xa(m),
     &          y01dp(n,2),y00dp(n,2),a0(2),b0(2),xmup(2)

      do 72 i = 1,n
         xx(i) = x1(i)
 72   continue
	
	one = 1.0D+00
	xmup(1) = dexp(a0(1)+a0(2)*dlog(xmu))
	xmup(2) = dexp(b0(1)+b0(2)*dlog(xmu))

      do 71 i = 1,m
	   xa(i) = x2(i)
 71   continue

      a(1) = 0.0D+00
      a(n) = -1.0D+00
      b(1) = 1.0D+00
      b(n) = 1.05D+00
      c(1) = -1.05D+00
      c(n) = 0.0D+00
      r(1) = 0.0D+00
      r(n) = 0.0D+00

      do 81 k = 1,4
	   do 79 jj = 1,2
            do 30 i = 1,n
	         do 29 j = 1,m
	            ya(j) = y(i,j,k)
 29            continue
               call polint(xa,ya,m,xmup(jj),yval,dy)
               yy(i) = yval
               if (k .eq. 1) y11(i,jj) = yy(i)
               if (k .eq. 2) y10(i,jj) = yy(i)
	         if (k .eq. 3) y01(i,jj) = yy(i)
	         if (k .eq. 4) y00(i,jj) = yy(i)
 30         continue

            do 21 i = 2,n-1
               a(i) = (xx(i)-xx(i-1))/6.0D+00
               b(i) = (xx(i+1)-xx(i-1))/3.0D+00
               c(i) = (xx(i+1)-xx(i))/6.0D+00
               r(i) = (yy(i+1)-yy(i))/(xx(i+1)-xx(i))-
     &                (yy(i)-yy(i-1))/(xx(i)-xx(i-1))
 21         continue

            call tridag(a,b,c,r,u,n)
            do 22 i = 1,n
               if (k .eq. 1) y11dp(i,jj) = u(i)
               if (k .eq. 2) y10dp(i,jj) = u(i)
	         if (k .eq. 3) y01dp(i,jj) = u(i)
	         if (k .eq. 4) y00dp(i,jj) = u(i)
 22         continue
 79      continue
 81   continue

      return
      end

      subroutine polint(xa,ya,n,x,y,dy)
	implicit real*8 (a-h,o-z)
	parameter (nmax=5,toler=1.0D-12)
	dimension xa(n),ya(n),c(nmax),d(nmax)

	ns = 1
	dif = dabs(x-xa(1))
	do 11 i = 1,n
	   dift = dabs(x-xa(i))
	   if (dift .lt. dif) then
	      ns = i
	      dif = dift
	   endif
	   c(i) = ya(i)
	   d(i) = ya(i)
 11   continue

      y = ya(ns)
	ns = ns-1
	do 13 m = 1,n-1
	   do 12 i = 1,n-m
	      ho = xa(i)-x
	      hp = xa(i+m)-x
	      w = c(i+1)-d(i)
	      den = ho-hp
	      if (dabs(den) .lt. toler) then
	         write(6,100)
 100           format('error in polint')
               pause
	      endif
	      den = w/den
	      d(i) = hp*den
	      c(i) = ho*den
 12      continue
         if ((2*ns) .lt. (n-m)) then
	      dy = c(ns+1)
	   else
	      dy = d(ns)
	      ns = ns-1
	   endif
         y = y+dy
 13   continue

      return
	end

      subroutine tridag(a,b,c,r,u,n)
      implicit real*8 (a-h,o-z)
      parameter (toler=1.0D-12,nmax=5000)
      dimension gam(nmax),a(n),b(n),c(n),r(n),u(n)

      bet = b(1)
      u(1) = r(1)/bet
      do 11 j = 2,n
         gam(j) = c(j-1)/bet
         bet = b(j)-a(j)*gam(j)
         if (dabs(bet) .lt. toler) then
            write(6,100)
 100        format('failure in tridag')
            pause
         endif
         u(j) = (r(j)-a(j)*u(j-1))/bet
 11   continue
      do 12 j = n-1,1,-1
         u(j) = u(j)-gam(j+1)*u(j+1)
 12   continue

      return
      end

      subroutine calopt(xk,neps,ynew,value,niter,toler,
     &                  r,w,ykhigh,yklow,grad,xkpts,
     &                  xmupts,nmu,nagg)
      implicit real*8 (a-h,o-z)
      parameter (mxiter=200,nkpts=150,hfix=1.0D+00/0.9D+00,
     &           delta=0.025D+00,beta=0.99D+00,nmupts=4,
     &           nprint=0,xmu=0.15D+00,gamma=1.0D+00,unempb=0.1D+00,
     &           unempg=0.04D+00)
      common/cprob/pi(4,4)
	common/cvs/v11(nkpts,nmupts,2),v11dp(nkpts,nmupts,2),
     &           v10(nkpts,nmupts,2),v10dp(nkpts,nmupts,2),
     &           v01(nkpts,nmupts,2),v01dp(nkpts,nmupts,2),
     &           v00(nkpts,nmupts,2),v00dp(nkpts,nmupts,2)
      dimension xkpts(nkpts),xmupts(nmupts),r(nmupts,2),
     &          w(nmupts,2)

      one = 1.0D+00
      niter = 0

      if (nagg .eq. 1) then
	   tau = xmu*unempg/((one-unempg)*hfix)
	else
	   tau = xmu*unempb/((one-unempb)*hfix)
	endif
      if (neps .eq. 1) then
	   y = r(nmu,nagg)*xk+(one-tau)*w(nmu,nagg)*hfix
      else
	   y = r(nmu,nagg)*xk+xmu*w(nmu,nagg)
      endif
      if (neps .eq. 1 .and. nagg .eq. 1) then
         npi = 1
      elseif (neps .eq. 0 .and. nagg .eq. 1) then
         npi = 2
	elseif (neps .eq. 1 .and. nagg .eq. 2) then
	   npi = 3
      elseif (neps .eq. 0 .and. nagg .eq. 2) then
	   npi = 4
      endif

      xklow = yklow
      xkhigh = ykhigh

      do 10 i = 1,mxiter
         xkcur = (xkhigh+xklow)/2.0D+00
         x = xkcur-(one-delta)*xk
         if (x .ge. y) then
            xkhigh = xkcur
            goto 10
         endif

         c = y-x
         upc = -c**(-gamma)

         call splint(xkpts,v11,v11dp,xkcur,yval,yp1,ydp,khi,klo,
     &               0,nkpts,0,1,0,nmu,nagg)
         call splint(xkpts,v10,v10dp,xkcur,yval,yp2,ydp,khi,klo,
     &               1,nkpts,0,1,0,nmu,nagg)
	   call splint(xkpts,v01,v01dp,xkcur,yval,yp3,ydp,khi,klo,
     &               1,nkpts,0,1,0,nmu,nagg)
	   call splint(xkpts,v00,v00dp,xkcur,yval,yp4,ydp,khi,klo,
     &               1,nkpts,0,1,0,nmu,nagg)
         
         grad = upc+beta*(pi(npi,1)*yp1+pi(npi,2)*yp2+
     &          pi(npi,3)*yp3+pi(npi,4)*yp4)

         if (grad .ge. 0.0D+00) then
            xklow = xkcur
         else
            xkhigh = xkcur
         endif

         if (nprint .eq. 1) then
	      write(6,955) xk,xmupts(nmu),neps,nagg,xkcur,grad
 955        format(2f12.7,2i3,2f18.14)
            write(6,757) yp1,yp2,yp3,yp4
 757        format(4f18.10)
            write(6,444) c
 444        format(f18.10)
            pause
	   endif

         if ((xkhigh-xklow) .le. toler) then
            niter = i
            goto 9
         endif

 10   continue

 9    xkcur = (xkhigh+xklow)/2.0D+00
      x = xkcur-(one-delta)*xk
      c = y-x

      if (c .le. 0.0D+00) then
         write(6,787) c
 787     format('failure in calopt: negative c',f15.10)
         pause
      endif

	if (dabs(gamma-one) .lt. 1.0D-04) then
	   utilc = dlog(c)
	else
	   utilc = one/(one-gamma)*(c**(one-gamma)-one)
	endif
      call splint(xkpts,v11,v11dp,xkcur,yval1,yp,ydp,khi,klo,
     &            0,nkpts,1,0,0,nmu,nagg)
      call splint(xkpts,v10,v10dp,xkcur,yval2,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nmu,nagg)
	call splint(xkpts,v01,v01dp,xkcur,yval3,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nmu,nagg)
	call splint(xkpts,v00,v00dp,xkcur,yval4,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nmu,nagg)

      value = utilc+beta*(pi(npi,1)*yval1+pi(npi,2)*yval2+
     &        pi(npi,3)*yval3+pi(npi,4)*yval4)
      ynew = xkcur

      return
      end

      subroutine calopt2(xk,neps,ynew,value,niter,toler,
     &                  r,w,ykhigh,yklow,grad,xkpts,
     &                  xmupts,xmmu,nagg)
      implicit real*8 (a-h,o-z)
      parameter (mxiter=200,nkpts=150,hfix=1.0D+00/0.9D+00,
     &           delta=0.025D+00,beta=0.99D+00,nmupts=4,
     &           nprint=0,xmu=0.15D+00,gamma=1.0D+00,unempg=0.04D+00,
     &           unempb=0.1D+00)
      common/cprob/pi(4,4)
	common/cws/w11(nkpts,2),w11dp(nkpts,2),
     &           w10(nkpts,2),w10dp(nkpts,2),
     &           w01(nkpts,2),w01dp(nkpts,2),
     &           w00(nkpts,2),w00dp(nkpts,2)
      dimension xkpts(nkpts),xmupts(nmupts),r(2),w(2)

      one = 1.0D+00
      niter = 0

      if (nagg .eq. 1) then
	   tau = xmu*unempg/((one-unempg)*hfix)
	else
	   tau = xmu*unempb/((one-unempb)*hfix)
	endif
      if (neps .eq. 1) then
	   y = r(nagg)*xk+(one-tau)*w(nagg)*hfix
      else
	   y = r(nagg)*xk+xmu*w(nagg)
      endif
      if (neps .eq. 1 .and. nagg .eq. 1) then
         npi = 1
      elseif (neps .eq. 0 .and. nagg .eq. 1) then
         npi = 2
	elseif (neps .eq. 1 .and. nagg .eq. 2) then
	   npi = 3
      elseif (neps .eq. 0 .and. nagg .eq. 2) then
	   npi = 4
      endif

      xklow = yklow
      xkhigh = ykhigh

      do 10 i = 1,mxiter
         xkcur = (xkhigh+xklow)/2.0D+00
         x = xkcur-(one-delta)*xk
         if (x .ge. y) then
            xkhigh = xkcur
            goto 10
         endif

         c = y-x
         upc = -c**(-gamma)

         call splin2(xkpts,w11,w11dp,xkcur,yval,yp1,ydp,khi,klo,
     &               0,nkpts,0,1,0,nagg)
         call splin2(xkpts,w10,w10dp,xkcur,yval,yp2,ydp,khi,klo,
     &               1,nkpts,0,1,0,nagg)
	   call splin2(xkpts,w01,w01dp,xkcur,yval,yp3,ydp,khi,klo,
     &               1,nkpts,0,1,0,nagg)
	   call splin2(xkpts,w00,w00dp,xkcur,yval,yp4,ydp,khi,klo,
     &               1,nkpts,0,1,0,nagg)
         
         grad = upc+beta*(pi(npi,1)*yp1+pi(npi,2)*yp2+
     &          pi(npi,3)*yp3+pi(npi,4)*yp4)

         if (grad .ge. 0.0D+00) then
            xklow = xkcur
         else
            xkhigh = xkcur
         endif

         if (nprint .eq. 1) then
	      write(6,955) xk,xmu,neps,nagg,grad,xkcur
 955        format(2f12.7,2i3,2f18.14)
            pause
	   endif

         if ((xkhigh-xklow) .le. toler) then
            niter = i
            goto 9
         endif

 10   continue

 9    xkcur = (xkhigh+xklow)/2.0D+00
      x = xkcur-(one-delta)*xk
      c = y-x

      if (c .le. 0.0D+00) then
         write(6,787) c
 787     format('failure in calopt: negative c',f15.10)
         pause
      endif

	if (dabs(gamma-one) .lt. 1.0D-04) then
	   utilc = dlog(c)
	else
	   utilc = one/(one-gamma)*(c**(one-gamma)-one)
	endif
      call splin2(xkpts,w11,w11dp,xkcur,yval1,yp,ydp,khi,klo,
     &            0,nkpts,1,0,0,nagg)
      call splin2(xkpts,w10,w10dp,xkcur,yval2,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nagg)
	call splin2(xkpts,w01,w01dp,xkcur,yval3,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nagg)
	call splin2(xkpts,w00,w00dp,xkcur,yval4,yp,ydp,khi,klo,
     &            1,nkpts,1,0,0,nagg)

      value = utilc+beta*(pi(npi,1)*yval1+pi(npi,2)*yval2+
     &        pi(npi,3)*yval3+pi(npi,4)*yval4)
      ynew = xkcur

      return
      end

      subroutine altopt(xk,neps,ynew,value,niter,toler,
     &                  startk,grad,xkpts,r,w,xmupts,nmu,
     &                  nagg,ndone)
      implicit real*8 (a-h,o-z)
      parameter (mxiter=200,hfix=1.0D+00/0.9D+00,nmupts=4,
     &           delta=0.025D+00,beta=0.99D+00,nkpts=150,
     &           xmu=0.15D+00,nprint=0,gamma=1.0D+00,unempb=0.1D+00,
     &           unempg=0.04D+00)
      common/cprob/pi(4,4)
	common/cvs/v11(nkpts,nmupts,2),v11dp(nkpts,nmupts,2),
     &           v10(nkpts,nmupts,2),v10dp(nkpts,nmupts,2),
     &           v01(nkpts,nmupts,2),v01dp(nkpts,nmupts,2),
     &           v00(nkpts,nmupts,2),v00dp(nkpts,nmupts,2)
      dimension xkpts(nkpts),xmupts(nmupts),r(nmupts,2),
     &          w(nmupts,2)

      one = 1.0D+00
      niter = 0
      ndone = 1

      if (nagg .eq. 1) then
	   tau = xmu*unempg/((one-unempg)*hfix)
	else
	   tau = xmu*unempb/((one-unempb)*hfix)
	endif
      if (neps .eq. 1) then
	   y = r(nmu,nagg)*xk+(one-tau)*w(nmu,nagg)*hfix
      else
	   y = r(nmu,nagg)*xk+xmu*w(nmu,nagg)
      endif
      if (neps .eq. 1 .and. nagg .eq. 1) then
         npi = 1
      elseif (neps .eq. 0 .and. nagg .eq. 1) then
         npi = 2
	elseif (neps .eq. 1 .and. nagg .eq. 2) then
	   npi = 3
      elseif (neps .eq. 0 .and. nagg .eq. 2) then
	   npi = 4
      endif
      xk2p = startk

      do 10 i = 1,mxiter
         xk2 = xk2p
         x = xk2-(one-delta)*xk
         c = y-x

         upc = -c**(-gamma)
         udpc = -gamma*c**(-gamma-one)

         call splint(xkpts,v11,v11dp,xk2,yval,yp1,ydp1,khi,klo,
     &               0,nkpts,0,1,1,nmu,nagg)
         call splint(xkpts,v10,v10dp,xk2,yval,yp2,ydp2,khi,klo,
     &               1,nkpts,0,1,1,nmu,nagg)
	   call splint(xkpts,v01,v01dp,xk2,yval,yp3,ydp3,khi,klo,
     &               1,nkpts,0,1,1,nmu,nagg)
	   call splint(xkpts,v00,v00dp,xk2,yval,yp4,ydp4,khi,klo,
     &               1,nkpts,0,1,1,nmu,nagg)
         
         grad = upc+beta*(pi(npi,1)*yp1+pi(npi,2)*yp2+
     &          pi(npi,3)*yp3+pi(npi,4)*yp4)         
         hess = udpc+beta*(pi(npi,1)*ydp1+pi(npi,2)*ydp2+
     &          pi(npi,3)*ydp3+pi(npi,4)*ydp4)

         xk2p = xk2-grad/hess

         if (nprint .eq. 1) then
	      write(6,797) xk,neps,nagg
	      write(6,788) xk2,xk2p,grad,hess
 797        format(f12.8,2i3)
 788        format(4f18.10)
            pause
         endif

         if (dabs(xk2p-xk2) .le. toler) then
            niter = i
            goto 9
         endif

 10   continue

 9    if (c .gt. 0.0D+00 .and. xk2 .ge. 0.0D+00 .and.
     &    dabs(grad) .lt. 1.0D-06) then

         call splint(xkpts,v11,v11dp,xk2,yval1,yp,ydp,khi,klo,
     &               0,nkpts,1,0,0,nmu,nagg)
         call splint(xkpts,v10,v10dp,xk2,yval2,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nmu,nagg)
	   call splint(xkpts,v01,v01dp,xk2,yval3,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nmu,nagg)
	   call splint(xkpts,v00,v00dp,xk2,yval4,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nmu,nagg)
         ynew = xk2
	   c = y+(one-delta)*xk-ynew
	   if (dabs(gamma-one) .lt. 1.0D-04) then
	      utilc = dlog(c)
	   else
	      utilc = one/(one-gamma)*(c**(one-gamma)-one)
	   endif
         value = utilc+beta*(pi(npi,1)*yval1+pi(npi,2)*yval2+
     &           pi(npi,3)*yval3+pi(npi,4)*yval4)
      else
         ndone = 0
      endif

      return
      end

      subroutine altopt2(xk,neps,ynew,value,niter,toler,
     &                   startk,grad,xkpts,r,w,xmupts,xmmu,
     &                   nagg,ndone)
      implicit real*8 (a-h,o-z)
      parameter (mxiter=200,hfix=1.0D+00/0.9D+00,nmupts=4,
     &           delta=0.025D+00,beta=0.99D+00,nkpts=150,
     &           xmu=0.15D+00,gamma=1.0D+00,unempb=0.1D+00,
     &           unempg=0.04D+00)
      common/cprob/pi(4,4)
	common/cws/w11(nkpts,2),w11dp(nkpts,2),
     &           w10(nkpts,2),w10dp(nkpts,2),
     &           w01(nkpts,2),w01dp(nkpts,2),
     &           w00(nkpts,2),w00dp(nkpts,2)
      dimension xkpts(nkpts),xmupts(nmupts),r(2),w(2)

      one = 1.0D+00
      niter = 0
      ndone = 1

      if (nagg .eq. 1) then
	   tau = xmu*unempg/((one-unempg)*hfix)
	else
	   tau = xmu*unempb/((one-unempb)*hfix)
	endif
      if (neps .eq. 1) then
	   y = r(nagg)*xk+(one-tau)*w(nagg)*hfix
      else
	   y = r(nagg)*xk+xmu*w(nagg)
      endif
      if (neps .eq. 1 .and. nagg .eq. 1) then
         npi = 1
      elseif (neps .eq. 0 .and. nagg .eq. 1) then
         npi = 2
	elseif (neps .eq. 1 .and. nagg .eq. 2) then
	   npi = 3
      elseif (neps .eq. 0 .and. nagg .eq. 2) then
	   npi = 4
      endif
      xk2p = startk

      do 10 i = 1,mxiter
         xk2 = xk2p
         x = xk2-(one-delta)*xk
         c = y-x

         upc = -c**(-gamma)
         udpc = -gamma*c**(-gamma-one)

         call splin2(xkpts,w11,w11dp,xk2,yval,yp1,ydp1,khi,klo,
     &               0,nkpts,0,1,1,nagg)
         call splin2(xkpts,w10,w10dp,xk2,yval,yp2,ydp2,khi,klo,
     &               1,nkpts,0,1,1,nagg)
	   call splin2(xkpts,w01,w01dp,xk2,yval,yp3,ydp3,khi,klo,
     &               1,nkpts,0,1,1,nagg)
	   call splin2(xkpts,w00,w00dp,xk2,yval,yp4,ydp4,khi,klo,
     &               1,nkpts,0,1,1,nagg)
         
         grad = upc+beta*(pi(npi,1)*yp1+pi(npi,2)*yp2+
     &          pi(npi,3)*yp3+pi(npi,4)*yp4)         
         hess = udpc+beta*(pi(npi,1)*ydp1+pi(npi,2)*ydp2+
     &          pi(npi,3)*ydp3+pi(npi,4)*ydp4)

         xk2p = xk2-grad/hess

         if (dabs(xk2p-xk2) .le. toler) then
            niter = i
            goto 9
         endif

 10   continue

 9    if (c .gt. 0.0D+00 .and. xk2 .ge. 0.0D+00 .and.
     &    dabs(grad) .lt. 1.0D-06) then
         call splin2(xkpts,w11,w11dp,xk2,yval1,yp,ydp,khi,klo,
     &               0,nkpts,1,0,0,nagg)
         call splin2(xkpts,w10,w10dp,xk2,yval2,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nagg)
	   call splin2(xkpts,w01,w01dp,xk2,yval3,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nagg)
	   call splin2(xkpts,w00,w00dp,xk2,yval4,yp,ydp,khi,klo,
     &               1,nkpts,1,0,0,nagg)
         ynew = xk2
	   c = y+(one-delta)*xk-ynew
	   if (dabs(gamma-one) .lt. 1.0D-04) then
	      utilc = dlog(c)
	   else
	      utilc = one/(one-gamma)*(c**(one-gamma)-one)
	   endif
         value = utilc+beta*(pi(npi,1)*yval1+pi(npi,2)*yval2+
     &           pi(npi,3)*yval3+pi(npi,4)*yval4)
      else
         ndone = 0
      endif

      return
      end

      subroutine hunt(xx,n,x,jlo)
      implicit real*8 (a-h,o-z)
      dimension xx(n)
      logical ascnd

      ascnd = xx(n) .ge. xx(1)
      if (jlo .le. 0 .or. jlo .gt. n) then
         jlo = 0
         jhi = n+1
         goto 3
      endif

      inc = 1
      
      if (x .ge. xx(jlo) .eqv. ascnd) then
 1       jhi = jlo+inc
         if (jhi .gt. n) then
            jhi = n+1
         elseif (x .ge. xx(jhi) .eqv. ascnd) then
            jlo = jhi
            inc = inc+inc
            goto 1
         endif
      else
         jhi = jlo
 2       jlo = jhi-inc
         if (jlo .lt. 1) then
            jlo = 0
         elseif (x .lt. xx(jlo) .eqv. ascnd) then
            jhi = jlo
            inc = inc+inc
            goto 2
         endif
      endif
 3    if (jhi-jlo .eq. 1) then
         if (x .eq. xx(n)) jlo = n-1
         if (x .eq. xx(1)) jlo = 1
         return
      endif
      jm = (jhi+jlo)/2
      if (x .ge. xx(jm) .eqv. ascnd) then
         jlo = jm
      else
         jhi = jm
      endif
      goto 3

      end

c     Simulation routine -- solves for equilibrium

      subroutine calcf(optkg1,optkg0,xk2pts,xm2pts,
     &                 a0,b0,a0new,b0new,xkmean,sdmean,
     &				 skmean,s4mean,s5mean,rsqa,rsqb,sderra,sderrb,
     &                 gmean,p1mean,p5mean,p10mean,p20mean,p30mean,
     &                 p90mean,p95mean,hansen,erfmx1,erfmx2)
      implicit real*8 (a-h,o-z)
      parameter (nk2pts=5000,ncapt=10000,ndrop=500,nm2pts=30,
     &           unempb=0.1D+00,unempg=0.04D+00,durub=2.5D+00,
     &           durug=1.5D+00,nread2=1,alpha=0.36D+00,xmu=0.15D+00,
     &           hfix=1.0D+00/0.9D+00,delta=0.025D+00,zgood=1.01D+00,
     &           zbad=0.99D+00,gamma=1.0D+00,beta=0.99D+00)
      common/cfs/f1(nk2pts,2)
      common/cprob/pi(4,4)
	common/nshock/nz(ncapt+1),nind(ncapt)
	common/cvgrid/vgrid(nk2pts,nm2pts,4)
	common/cgmm/rent(ncapt),cons(ncapt)
	common/vgmm/vhat(3,3),wght(3,3)
	external gmm
      dimension a0(2),b0(2),a0new(2),b0new(2),atinv(28,28),
     &          optkg1(nk2pts,nm2pts,2),abdev(ncapt),temp(1,28),
     &          optkg0(nk2pts,nm2pts,2),fracs(ncapt),btt(1,28),
     &          xk2pts(nk2pts),xm2pts(nm2pts),xdata3(ncapt,28),
     &          xkvals(ncapt+1),f0(nk2pts,2),xdata2(ncapt,28),
     &          xdata(ncapt,2),ydata(ncapt),stdxk(ncapt),ttt(1,1),
     &          skewxk(ncapt),xkurtk(ncapt),ginis(ncapt),
     &          pct1(ncapt),pct5(ncapt),pct10(ncapt),pct20(ncapt),
     &          pct30(ncapt),toppct(7),tops(7),bt(28,1),at(28,28),
     &          vmus(ncapt),xdatan(ncapt,6),ydatan(ncapt),bhat(6),
     &          xdatas(ncapt,8),p0new(8),q0new(8),p(3,2),y(3),
     &          xkind(ncapt+1),cind(ncapt),xkfive(ncapt),pct90(ncapt),
     &          pct95(ncapt),xkpred(ncapt+1),xkbare(ncapt),
     &          xkbaru(ncapt),wage(ncapt),gdp(ncapt),
     &          yind(ncapt),cgrow(ncapt)

      one = 1.0D+00
	xkinc = xk2pts(2)-xk2pts(1)
      xminc = xm2pts(2)-xm2pts(1)
	toppct(1) = 0.01D+00
	toppct(2) = 0.05D+00
	toppct(3) = 0.1D+00
	toppct(4) = 0.2D+00
	toppct(5) = 0.3D+00
	toppct(6) = 0.9D+00
	toppct(7) = 0.95D+00

      pgg00 = (durug-one)/durug
	pbb00 = (durub-one)/durub
	pbg00 = 1.25D+00*pbb00
	pgb00 = 0.75D+00*pgg00
	pgg01 = (unempg-unempg*pgg00)/(one-unempg)
	pbb01 = (unempb-unempb*pbb00)/(one-unempb)
	pbg01 = (unempb-unempg*pbg00)/(one-unempg)
	pgb01 = (unempg-unempb*pgb00)/(one-unempb)

	pgg10 = one-(durug-one)/durug
	pbb10 = one-(durub-one)/durub
	pbg10 = one-1.25D+00*pbb00
	pgb10 = one-0.75D+00*pgg00
	pgg11 = one-(unempg-unempg*pgg00)/(one-unempg)
	pbb11 = one-(unempb-unempb*pbb00)/(one-unempb)
	pbg11 = one-(unempb-unempg*pbg00)/(one-unempg)
	pgb11 = one-(unempg-unempb*pgb00)/(one-unempb)

      xkmean = 0.0D+00
	sdmean = 0.0D+00
	skmean = 0.0D+00
	s4mean = 0.0D+00
	s5mean = 0.0D+00
	gmean = 0.0D+00
	p1mean = 0.0D+00
	p5mean = 0.0D+00
	p10mean = 0.0D+00
	p20mean = 0.0D+00
	p30mean = 0.0D+00
	p90mean = 0.0D+00
	p95mean = 0.0D+00
	
	contot = 0.0D+00
	congd = 0.0D+00
	conbd = 0.0D+00
	mgood = 0
	mbad = 0

	xkmne = 0.0D+00
	xkmnu = 0.0D+00

      open(2,file='saving_final.f1',form='formatted')
      do 66 i = 1,nk2pts
         read(2,547) dum,f1(i,1),f1(i,2)
 547     format(3f18.10)
 66   continue
      close(2)
	fsum = 0.0D+00
	do 38 i = 1,nk2pts
          fsum = fsum+f1(i,1)+f1(i,2)
 38   continue
      if (dabs(fsum-one) .gt. 1.0D-04) then
	   write(6,575) fsum
 575     format('initial distribution integrates to :',f18.10)
         pause
	endif

      xkbcur = 0.0D+00
	do 45 i = 1,nk2pts
	   xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
 45   continue

      xkind(1) = xkbcur
	xkpred(1) = xkbcur

      open(7,file='jedc_final_1.sim',form='formatted')
	open(8,file='jedc_final_1.ind',form='formatted')
      do 12 ii = 1,ncapt

         nzcur = nz(ii)
         if (nzcur .eq. 1) then
	      nz2 = 1
	      nz3 = 2
	   else
	      nz2 = 3
	      nz3 = 4
	   endif
	   if (nzcur .eq. 1) then   
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgg11
	         prob1 = pgg10
	      else
	         prob0 = pbg11
	         prob1 = pbg10
	      endif
	   else
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgb11
	         prob1 = pgb10
	      else
	         prob0 = pbb11
	         prob1 = pbb10
	      endif
         endif
         xkvals(ii) = xkbcur

         if (nz(ii) .eq. 1) then
	      rent(ii) = alpha*zgood*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempg))**(one-alpha)-delta
	      wage(ii) = (one-alpha)*zgood*xkvals(ii)**alpha*
     &                 (hfix*(one-unempg))**(-alpha)
	      gdp(ii) = zgood*xkvals(ii)**alpha*(hfix*(one-unempg))**
     &                (one-alpha)
	      tau = xmu*unempg/((one-unempg)*hfix)
	   else
	      rent(ii) = alpha*zbad*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempb))**(one-alpha)-delta
	      wage(ii) = (one-alpha)*zbad*xkvals(ii)**alpha*
     &                 (hfix*(one-unempb))**(-alpha)
	      gdp(ii) = zbad*xkvals(ii)**alpha*(hfix*(one-unempb))**
     &                (one-alpha)
	      tau = xmu*unempb/((one-unempb)*hfix)
	   endif

         f0 = f1
         f1 = 0.0D+00

         call hunt(xm2pts,nm2pts,xkbcur,mcor)
         if (mcor .lt. 1) then
            capa = 0.0D+00
            mcor = 1
         elseif (mcor .ge. nm2pts) then
            capa = 1.0D+00
            mcor = nm2pts-1
         else
            capa = (xkbcur-xm2pts(mcor))/xminc
         endif

         vmu = 0.0D+00
         do 88 i = 1,nk2pts

            xk1 = optkg1(i,mcor,nzcur)+capa*
     &            (optkg1(i,mcor+1,nzcur)-optkg1(i,mcor,nzcur))
            xk0 = optkg0(i,mcor,1)+capa*
     &            (optkg0(i,mcor+1,nzcur)-optkg0(i,mcor,nzcur))
	      v1 = vgrid(i,mcor,nz2)+capa*
     &           (vgrid(i,mcor+1,nz2)-vgrid(i,mcor,nz2))
	      v0 = vgrid(i,mcor,nz3)+capa*
     &           (vgrid(i,mcor+1,nz3)-vgrid(i,mcor,nz3))
            
            call hunt(xk2pts,nk2pts,xk1,jlo1)
            if (jlo1 .lt. 1) then
               wght1 = 1.0D+00
               jlo1 = 1
            elseif (jlo1 .ge. nk2pts) then
               wght1 = 0.0D+00
               jlo1 = nk2pts-1
            else
               wght1 = one-(xk1-xk2pts(jlo1))/xkinc
            endif

            call hunt(xk2pts,nk2pts,xk0,jlo0)
            if (jlo0 .lt. 1) then
               wght0 = 1.0D+00
               jlo0 = 1
            elseif (jlo0 .ge. nk2pts) then
               wght0 = 0.0D+00
               jlo0 = nk2pts-1
            else
               wght0 = one-(xk0-xk2pts(jlo0))/xkinc
            endif

            f1(jlo1,1) = prob0*wght1*f0(i,1)+f1(jlo1,1)
            f1(jlo1+1,1) = prob0*(one-wght1)*f0(i,1)+
     &                     f1(jlo1+1,1)
            f1(jlo0,1) = prob1*wght0*f0(i,2)+f1(jlo0,1)
            f1(jlo0+1,1) = prob1*(one-wght0)*f0(i,2)+
     &                     f1(jlo0+1,1)
            
		  f1(jlo1,2) = (one-prob0)*wght1*f0(i,1)+f1(jlo1,2)
            f1(jlo1+1,2) = (one-prob0)*(one-wght1)*
     &                      f0(i,1)+f1(jlo1+1,2)
            f1(jlo0,2) = (one-prob1)*wght0*f0(i,2)+f1(jlo0,2)
            f1(jlo0+1,2) = (one-prob1)*(one-wght0)*
     &                      f0(i,2)+f1(jlo0+1,2)

            vmu = vmu+v1*f0(i,1)+v0*f0(i,2)

 88      continue
         vmus(ii) = vmu

         call hunt(xk2pts,nk2pts,xkind(ii),jk)
	   if (jk .lt. 1) then
	      jk = 1
	      capk = 0.0D+00
	   elseif (jk .ge. nk2pts) then
	      jk = nk2pts-1
	      capk = 1.0D+00
	   else
	      capk = (xkind(ii)-xk2pts(jk))/xkinc
	   endif
	   if (nind(ii) .eq. 1) then
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg1(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg1(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg1(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg1(jk+1,mcor+1,nzcur)
	      yind(ii) = (one+rent(ii))*xkind(ii)+(one-tau)*wage(ii)*hfix
	      cind(ii) = yind(ii)-xkind(ii+1)
	   else
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg0(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg0(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg0(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg0(jk+1,mcor+1,nzcur)
	      yind(ii) = (one+rent(ii))*xkind(ii)+xmu*wage(ii)
	      cind(ii) = yind(ii)-xkind(ii+1)
	   endif

         xkbcur = 0.0D+00
	   fsum = 0.0D+00
         do 83 i = 1,nk2pts
            xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
	      fsum = fsum+f1(i,1)+f1(i,2)
 83      continue
         if (dabs(fsum-one) .gt. 1.0D-04) then
	      write(6,783) ii,fsum
 783        format('does not integrate to one in period ',i6,f18.10)
            pause
	   endif
	   if (nz(ii) .eq. 1) then
	      xkpred(ii+1) = dexp(a0(1)+a0(2)*dlog(xkpred(ii)))
	   else
	      xkpred(ii+1) = dexp(b0(1)+b0(2)*dlog(xkpred(ii)))
	   endif
         xinv = xkbcur-(one-delta)*xkvals(ii)
	   cons(ii) = gdp(ii)-xinv

         varcur = 0.0D+00
         skwcur = 0.0D+00
         xkurt = 0.0D+00
	   unemp = 0.0D+00
	   frpoor = 0.0D+00
	   absval = 0.0D+00
	   fiver = 0.0D+00
	   xke = 0.0D+00
	   xku = 0.0D+00
         do 82 i = 1,nk2pts
	      fterm = f0(i,1)+f0(i,2)
	      xterm = xk2pts(i)-xkvals(ii)
            varcur = varcur+fterm*xterm*xterm
            skwcur = skwcur+fterm*xterm*xterm*xterm
            xkurt = xkurt+fterm*xterm*xterm*xterm*xterm
	      fiver = fiver+fterm*xterm*xterm*xterm*xterm*xterm
	      unemp = unemp+f0(i,2)
	      absval = absval+dabs(xterm)*fterm
	      if (xk2pts(i) .lt. 0.1D+00*xkvals(ii)) frpoor = frpoor+fterm
	      xke = xke+xk2pts(i)*f0(i,1)
	      xku = xku+xk2pts(i)*f0(i,2)
 82      continue
         stdxk(ii) = dsqrt(varcur)/xkvals(ii)
         skewxk(ii) = skwcur**(one/3.0D+00)/xkvals(ii)
         xkurtk(ii) = xkurt**(one/4.0D+00)/xkvals(ii)
	   xkfive(ii) = fiver**(one/5.0D+00)/xkvals(ii)
	   fracs(ii) = frpoor
	   abdev(ii) = absval
	   xkbare(ii) = xke/(one-unemp)
	   xkbaru(ii) = xku/unemp

         call getgini(f0,xk2pts,xkvals(ii),ginis(ii),toppct,tops)
         pct1(ii) = tops(1)
         pct5(ii) = tops(2)
	   pct10(ii) = tops(3)
	   pct20(ii) = tops(4)
	   pct30(ii) = tops(5)
	   pct90(ii) = tops(6)
	   pct95(ii) = tops(7)

	   write(7,504) ii,nz(ii),xkvals(ii),
     &                stdxk(ii),skewxk(ii),xkurtk(ii),xkfive(ii),unemp,
     &                fracs(ii),abdev(ii),rent(ii),wage(ii),ginis(ii),
     &                gdp(ii),cons(ii),xinv,pct1(ii),pct5(ii),pct10(ii),
     &                pct20(ii),pct30(ii),pct90(ii),pct95(ii),vmus(ii),
     &                xkpred(ii),xkbare(ii),xkbaru(ii)
 504     format(i6,i3,25f15.9)
         write(8,556) ii,nz(ii),nind(ii),xkind(ii),cind(ii),yind,
     &                cons(ii),gdp(ii),xkvals(ii)
 556     format(i6,2i3,6f15.9)

         if (ii .gt. ndrop) then
            xkmean = xkmean+xkvals(ii)/dble(real(ncapt-ndrop-1))
            sdmean = sdmean+stdxk(ii)/dble(real(ncapt-ndrop-1))
            skmean = skmean+skewxk(ii)/dble(real(ncapt-ndrop-1))
            s4mean = s4mean+xkurtk(ii)/dble(real(ncapt-ndrop-1))
	      s5mean = s5mean+xkfive(ii)/dble(real(ncapt-ndrop-1))
	      gmean = gmean+ginis(ii)/dble(real(ncapt-ndrop-1))
	      p1mean = p1mean+pct1(ii)/dble(real(ncapt-ndrop-1))
	      p5mean = p5mean+pct5(ii)/dble(real(ncapt-ndrop-1))
	      p10mean = p10mean+pct10(ii)/dble(real(ncapt-ndrop-1))
	      p20mean = p20mean+pct20(ii)/dble(real(ncapt-ndrop-1))
	      p30mean = p30mean+pct30(ii)/dble(real(ncapt-ndrop-1))
	      p90mean = p90mean+pct90(ii)/dble(real(ncapt-ndrop-1))
	      p95mean = p95mean+pct95(ii)/dble(real(ncapt-ndrop-1))
	      contot = contot+(f0(1,1)+f0(1,2))/dble(real(ncapt-ndrop-1))
	      if (nz(ii) .eq. 1) then
	         congd = congd+(f0(1,1)+f0(1,2))
	         mgood = mgood+1
	      else
	         conbd = conbd+(f0(1,1)+f0(1,2))
	         mbad = mbad+1
	      endif
         endif

 12   continue
      close(7)
	close(8)

	congd = congd/dble(real(mgood))
	conbd = conbd/dble(real(mbad))
	write(6,713) contot,congd,conbd
 713  format('const: ',3f18.10)

      cgrow(1) = 0.0D+00
	do 719 i = 2,ncapt
	   cgrow(i) = cons(i)/cons(i-1)
 719  continue

      rhocic = correl(cind(ndrop+1:ncapt),cons(ndrop+1:ncapt),
     &                ncapt-ndrop)
	rhociy = correl(cind(ndrop+1:ncapt),gdp(ndrop+1:ncapt),
     &                ncapt-ndrop)
      rhocik = correl(cind(ndrop+1:ncapt),xkvals(ndrop+1:ncapt),
     &                ncapt-ndrop)
	rhociyi = correl(cind(ndrop+1:ncapt),yind(ndrop+1:ncapt),
     &                 ncapt-ndrop)
      rhociki = correl(cind(ndrop+1:ncapt),xkind(ndrop+1:ncapt),
     &                 ncapt-ndrop)
	sigci = dstd(cind(ndrop+1:ncapt),ncapt-ndrop)
	sigki = dstd(xkind(ndrop+1:ncapt),ncapt-ndrop)
	rhocici1 = correl(cind(ndrop+1:ncapt),cind(ndrop:ncapt-1),
     &                  ncapt-ndrop)
	rhocici2 = correl(cind(ndrop+1:ncapt),cind(ndrop-1:ncapt-2),
     &                  ncapt-ndrop)
      rhocici3 = correl(cind(ndrop+1:ncapt),cind(ndrop-2:ncapt-3),
     &                  ncapt-ndrop)
	rhokiki1 = correl(xkind(ndrop+1:ncapt),xkind(ndrop:ncapt-1),
     &                  ncapt-ndrop)
	rhokiki2 = correl(xkind(ndrop+1:ncapt),xkind(ndrop-1:ncapt-2),
     &                  ncapt-ndrop)
      rhokiki3 = correl(xkind(ndrop+1:ncapt),xkind(ndrop-2:ncapt-3),
     &                  ncapt-ndrop)
      rhocg = correl(cgrow(ndrop+1:ncapt),cgrow(ndrop:ncapt-1),
     &               ncapt-ndrop)

      npts = 0
	do 34 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdata(npts,1) = one
	      xdata(npts,2) = dlog(xkvals(i))
            ydata(npts) = dlog(xkvals(i+1))
	   endif
 34   continue
      call ols2(ydata,xdata,ncapt,a0new,rsqa,sderra,npts)

	npts = 0
	do 33 i = ndrop+1,ncapt-1
         if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdata(npts,1) = one
	      xdata(npts,2) = dlog(xkvals(i))
	      ydata(npts) = dlog(xkvals(i+1))
	   endif
 33   continue
      call ols2(ydata,xdata,ncapt,b0new,rsqb,sderrb,npts)

      npts = 0
	do 39 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdata2(npts,1) = one
	      xdata2(npts,2) = dlog(xkvals(i))
	      xdata2(npts,3) = xdata2(npts,2)*xdata2(npts,2)
	      xdata2(npts,4) = dlog(stdxk(i))
	      xdata2(npts,5) = xdata2(npts,4)*xdata2(npts,4)
	      xdata2(npts,6) = xdata2(npts,2)*xdata2(npts,4)
	      xdata2(npts,7) = skewxk(i)
	      xdata2(npts,8) = xdata2(npts,7)*xdata2(npts,7)
	      xdata2(npts,9) = dlog(xkurtk(i))
	      xdata2(npts,10) = xdata2(npts,9)*xdata2(npts,9)
	      xdata2(npts,11) = xdata2(npts,2)*xdata2(npts,7)
	      xdata2(npts,12) = xdata2(npts,2)*xdata2(npts,9)
	      xdata2(npts,13) = xdata2(npts,4)*xdata2(npts,7)
	      xdata2(npts,14) = xdata2(npts,4)*xdata2(npts,9)
	      xdata2(npts,15) = xdata2(npts,7)*xdata2(npts,9)
	      xdata2(npts,16) = dlog(fracs(i))
	      xdata2(npts,17) = xdata2(npts,16)*xdata2(npts,16)
	      xdata2(npts,18) = xdata2(npts,16)*xdata2(npts,2)
	      xdata2(npts,19) = xdata2(npts,16)*xdata2(npts,4)
	      xdata2(npts,20) = xdata2(npts,16)*xdata2(npts,7)
	      xdata2(npts,21) = xdata2(npts,16)*xdata2(npts,9)
	      xdata2(npts,22) = dlog(abdev(i))
	      xdata2(npts,23) = xdata2(npts,22)*xdata2(npts,22)
	      xdata2(npts,24) = xdata2(npts,22)*xdata2(npts,2)	      
	      xdata2(npts,25) = xdata2(npts,22)*xdata2(npts,4)
	      xdata2(npts,26) = xdata2(npts,22)*xdata2(npts,7)
	      xdata2(npts,27) = xdata2(npts,22)*xdata2(npts,9)
	      xdata2(npts,28) = xdata2(npts,22)*xdata2(npts,16)
         endif
 39   continue

      npts = 0
	do 98 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdata3(npts,1) = one
	      xdata3(npts,2) = dlog(xkvals(i))
	      xdata3(npts,3) = xdata3(npts,2)*xdata3(npts,2)
	      xdata3(npts,4) = dlog(stdxk(i))
	      xdata3(npts,5) = xdata3(npts,4)*xdata3(npts,4)
	      xdata3(npts,6) = xdata3(npts,2)*xdata3(npts,4)
	      xdata3(npts,7) = skewxk(i)
	      xdata3(npts,8) = xdata3(npts,7)*xdata3(npts,7)
	      xdata3(npts,9) = dlog(xkurtk(i))
	      xdata3(npts,10) = xdata3(npts,9)*xdata3(npts,9)
	      xdata3(npts,11) = xdata3(npts,2)*xdata3(npts,7)
	      xdata3(npts,12) = xdata3(npts,2)*xdata3(npts,9)
	      xdata3(npts,13) = xdata3(npts,4)*xdata3(npts,7)
	      xdata3(npts,14) = xdata3(npts,4)*xdata3(npts,9)
	      xdata3(npts,15) = xdata3(npts,7)*xdata3(npts,9)
	      xdata3(npts,16) = dlog(fracs(i))
	      xdata3(npts,17) = xdata3(npts,16)*xdata3(npts,16)
	      xdata3(npts,18) = xdata3(npts,16)*xdata3(npts,2)
	      xdata3(npts,19) = xdata3(npts,16)*xdata3(npts,4)
	      xdata3(npts,20) = xdata3(npts,16)*xdata3(npts,7)
	      xdata3(npts,21) = xdata3(npts,16)*xdata3(npts,9)
	      xdata3(npts,22) = dlog(abdev(i))
	      xdata3(npts,23) = xdata3(npts,22)*xdata3(npts,22)
	      xdata3(npts,24) = xdata3(npts,22)*xdata3(npts,2)	      
	      xdata3(npts,25) = xdata3(npts,22)*xdata3(npts,4)
	      xdata3(npts,26) = xdata3(npts,22)*xdata3(npts,7)
	      xdata3(npts,27) = xdata3(npts,22)*xdata3(npts,9)
	      xdata3(npts,28) = xdata3(npts,22)*xdata3(npts,16)
         endif
 98   continue

	nz1 = 0
	nz0 = 0
	bt = 0.0D+00
	at = 0.0D+00
	capt = dble(real(ncapt-ndrop-1))
	erfmx1 = 0.0D+00
      erfmx2 = 0.0D+00
	do 78 i = ndrop,ncapt-1
	   if (nz(i) .eq. 1) then
	      nz1 = nz1+1
	      error = dlog(xkvals(i+1))-a0new(1)-
     &              a0new(2)*dlog(xkvals(i))
	      do 77 j = 1,28
	         bt(j,1) = bt(j,1)+error*xdata2(nz1,j)/capt
	         btt(1,j) = bt(j,1)
               do 76 k = 1,28
	            at(j,k) = at(j,k)+error*error*xdata2(nz1,j)*
     &                      xdata2(nz1,k)/capt
 76            continue
 77         continue
            if (dabs(error) .ge. erfmx1) erfmx1 = dabs(error)
         else
	      nz0 = nz0+1
	      error = dlog(xkvals(i+1))-b0new(1)-
     &              b0new(2)*dlog(xkvals(i))
	      do 79 j = 1,28
	         bt(j,1) = bt(j,1)+error*xdata3(nz0,j)/capt
	         btt(1,j) = bt(j,1)
	         do 80 k = 1,28
	            at(j,k) = at(j,k)+error*error*xdata3(nz0,j)*
     &                      xdata3(nz0,k)/capt
 80            continue
 79         continue
            if (dabs(error) .ge. erfmx2) erfmx2 = dabs(error)
         endif
 78   continue
      call inv28(at,atinv,nsing)
	temp = matmul(btt,atinv)
	ttt = matmul(temp,bt)
	hansen = ttt(1,1)*capt

      npts = 0
	do 70 i = ndrop,ncapt-1
	   npts = npts+1
	   if (nz(i) .eq. 1) then
            error = dlog(xkvals(i+1))-a0new(1)-a0new(2)*dlog(xkvals(i))
	   else
	      error = dlog(xkvals(i+1))-b0new(1)-b0new(2)*dlog(xkvals(i))
	   endif
	   xdatan(npts,1) = dlog(abdev(i))
	   xdatan(npts,2) = dlog(stdxk(i))
	   xdatan(npts,3) = dlog(skewxk(i))
	   xdatan(npts,4) = dlog(xkurtk(i))
	   xdatan(npts,5) = dlog(ginis(i))
	   xdatan(npts,6) = dlog(fracs(i))
	   ydatan(npts) = error
 70   continue
      call ols6(ydatan,xdatan,ncapt,bhat,rsqe,sderre,npts)

      write(6,289) 
 289  format('naive test statistics')
	write(6,729) (bhat(i),i=1,6)
 729  format(3f18.10)
      write(6,729) rsqe,sderre

      npts = 0
	do 354 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdatas(npts,1) = one
	      xdatas(npts,2) = dlog(xkvals(i))
	      xdatas(npts,3) = dlog(stdxk(i))
	      xdatas(npts,4) = dlog(skewxk(i))
	      xdatas(npts,5) = dlog(xkurtk(i))
	      xdatas(npts,6) = dlog(ginis(i))
	      xdatas(npts,7) = dlog(fracs(i))
	      xdatas(npts,8) = dlog(abdev(i))
            ydata(npts) = dlog(stdxk(i+1))
	   endif
 354  continue
      call ols8(ydata,xdatas,ncapt,p0new,rsqp,sderrp,npts)

	npts = 0
	do 353 i = ndrop+1,ncapt-1
         if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdatas(npts,1) = one
	      xdatas(npts,2) = dlog(xkvals(i))
	      xdatas(npts,3) = dlog(stdxk(i))
	      xdatas(npts,4) = dlog(skewxk(i))
	      xdatas(npts,5) = dlog(xkurtk(i))
	      xdatas(npts,6) = dlog(ginis(i))
	      xdatas(npts,7) = dlog(fracs(i))
	      xdatas(npts,8) = dlog(abdev(i))
            ydata(npts) = dlog(stdxk(i+1))
	   endif
 353  continue
      call ols8(ydata,xdatas,ncapt,q0new,rsqq,sderrq,npts)

      write(6,299) 
 299  format('naive test statistics for Sigma')
	write(6,739) (p0new(i),i=1,4)
 739  format(4f18.10)
	write(6,739) (p0new(i),i=5,8)
      write(6,729) rsqp,sderrp      
	write(6,739) (q0new(i),i=1,4)
	write(6,739) (q0new(i),i=5,8)
      write(6,729) rsqq,sderrq      

      vhat(1,1) = one
	vhat(1,2) = 0.0D+00
	vhat(1,3) = 0.0D+00
	vhat(2,1) = 0.0D+00
	vhat(2,2) = one
	vhat(2,3) = 0.0D+00
	vhat(1,3) = 0.0D+00
	vhat(2,3) = 0.0D+00
	vhat(3,3) = one
      call inv3(vhat,wght,nsing)

      p(1,1) = beta
	p(1,2) = gamma
	p(2,1) = 0.8D+00
	p(2,2) = gamma
	p(3,1) = beta
	p(3,2) = 4.0D+00
	do 67 i = 1,3
	   y(i) = gmm(p(i,:))
 67   continue
	call amoeba(p,y,3,2,2,1.0D-08,gmm,niter)
	bhat2 = p(1,1)
	ghat2 = p(1,2)

      vhat = 0.0D+00
      do 56 i = ndrop+1,ncapt-5
	   err1 = one-bhat2*(cons(i+1)/cons(i))**(-ghat2)*(one+rent(i+1))
	   err2 = err1*rent(i)
	   err3 = err1*cons(i)/cons(i-1)
	   errm11 = one-bhat2*(cons(i)/cons(i-1))**(-ghat2)*(one+rent(i))
	   errm21 = one-bhat2*(cons(i-1)/cons(i-2))**(-ghat2)*
     &            (one+rent(i-1))
	   errm31 = one-bhat2*(cons(i-2)/cons(i-3))**(-ghat2)*
     &            (one+rent(i-2))
	   errp11 = one-bhat2*(cons(i+2)/cons(i+1))**(-ghat2)*
     &            (one+rent(i+1))
	   errp21 = one-bhat2*(cons(i+3)/cons(i+2))**(-ghat2)*
     &            (one+rent(i+2))
	   errp31 = one-bhat2*(cons(i+4)/cons(i+3))**(-ghat2)*
     &            (one+rent(i+3))
	   errm12 = errm11*rent(i-1)
	   errm22 = errm21*rent(i-2)
	   errm32 = errm31*rent(i-3)
	   errp12 = errp11*rent(i+1)
	   errp22 = errp21*rent(i+2)
	   errp32 = errp31*rent(i+3)
	   errm13 = errm13*cons(i-1)/cons(i-2)
	   errm23 = errm23*cons(i-2)/cons(i-3)
	   errm33 = errm33*cons(i-3)/cons(i-4)
	   errp13 = errp13*cons(i+1)/cons(i)
	   errp23 = errp23*cons(i+2)/cons(i+1)
	   errp33 = errp33*cons(i+3)/cons(i+2)

         vhat(1,1) = vhat(1,1)+err1*err1+
     &               0.75D+00*(errm11*errm11+errp11*errp11)+
     &               0.50D+00*(errm21*errm21+errp21*errp21)+
     &               0.25D+00*(errm31*errm31+errp31*errp31)
	   vhat(1,2) = vhat(1,2)+err1*err2+
     &               0.75D+00*(errm11*errm12+errp11*errp12)+
     &               0.50D+00*(errm21*errm22+errp21*errp22)+
     &               0.25D+00*(errm31*errm32+errp31*errp32)
	   vhat(1,3) = vhat(1,3)+err1*err3+
     &               0.75D+00*(errm11*errm13+errp11*errp13)+
     &               0.50D+00*(errm21*errm23+errp21*errp23)+
     &               0.25D+00*(errm31*errm33+errp31*errp33)
	   vhat(2,1) = vhat(2,1)+err2*err1+
     &               0.75D+00*(errm12*errm11+errp12*errp11)+
     &               0.50D+00*(errm22*errm21+errp22*errp21)+
     &               0.25D+00*(errm32*errm31+errp32*errp31)
	   vhat(2,2) = vhat(2,2)+err2*err2+
     &               0.75D+00*(errm12*errm12+errp12*errp12)+
     &               0.50D+00*(errm22*errm22+errp22*errp22)+
     &               0.25D+00*(errm32*errm32+errp32*errp32)
	   vhat(2,3) = vhat(2,3)+err2*err3+
     &               0.75D+00*(errm12*errm13+errp12*errp13)+
     &               0.50D+00*(errm22*errm23+errp22*errp23)+
     &               0.25D+00*(errm32*errm33+errp32*errp33)
	   vhat(3,1) = vhat(3,1)+err3*err1+
     &               0.75D+00*(errm13*errm11+errp13*errp11)+
     &               0.50D+00*(errm23*errm21+errp23*errp21)+
     &               0.25D+00*(errm33*errm31+errp33*errp31)
	   vhat(3,2) = vhat(3,2)+err3*err2+
     &               0.75D+00*(errm13*errm12+errp13*errp12)+
     &               0.50D+00*(errm23*errm22+errp23*errp22)+
     &               0.25D+00*(errm33*errm32+errp33*errp32)
	   vhat(3,3) = vhat(3,3)+err3*err3+
     &               0.75D+00*(errm13*errm13+errp13*errp13)+
     &               0.50D+00*(errm23*errm23+errp23*errp23)+
     &               0.25D+00*(errm33*errm33+errp33*errp33)
 56   continue
      call inv3(vhat,wght,nsing)

      p(1,1) = bhat2
	p(1,2) = ghat2
	p(2,1) = 0.8D+00
	p(2,2) = ghat2
	p(3,1) = bhat2
	p(3,2) = 4.0D+00
	do 60 i = 1,3
	   y(i) = gmm(p(i,:))
 60   continue
	call amoeba(p,y,3,2,2,1.0D-08,gmm,niter)
	bhat2 = p(1,1)
	ghat2 = p(1,2)

      write(6,974)
 974  format('GMM estimates')
      write(6,737) beta,bhat2,gamma,ghat2
	write(6,737) y(1)
 737  format(4f18.10)

      return
      end

      subroutine getgini(f1,xk2pts,xkmean,gini,toppct,tops)
	implicit real*8 (a-h,o-z)
	parameter (nk2pts=5000)
	dimension xpt(nk2pts),ypt(nk2pts),aa(nk2pts),bb(nk2pts),
     &          cc(nk2pts),rr(nk2pts),ydp(nk2pts),xk2pts(nk2pts),
     &          f1(nk2pts,2),tops(7),toppct(7)

      one = 1.0D+00

      sumx = 0.0D+00
      sumy = 0.0D+00
      xpt(1) = 0.0D+00
      ypt(1) = 0.0D+00
      nhpts = 1
      do 89 j = 2,nk2pts
         sumx = sumx+f1(j-1,1)+f1(j-1,2)
         sumy = sumy+(f1(j-1,1)+f1(j-1,2))*xk2pts(j-1)/
     &          xkmean
         if (f1(j-1,1)+f1(j-1,2) .gt. 1.0D-10) then
            nhpts = nhpts+1
            xpt(nhpts) = sumx
            ypt(nhpts) = sumy
         endif
 89   continue

      call calcsd(xpt,ypt,ydp,aa,bb,cc,rr,nhpts,-one,-one)
      call integr(xpt,ypt,ydp,nhpts,xint)
      gini = one-2.0D+00*xint

      do 10 i = 1,7
	   call spling(xpt,ypt,ydp,one-toppct(i),yval,yp,ydp,khi,klo,
     &               0,nhpts,1,0,0)
	   tops(i) = one-yval
 10   continue

      return
      end

      subroutine integr(xa,ya,y2a,n,xint)
      implicit real*8 (a-h,o-z)
      dimension xa(n),ya(n),y2a(n)

      one = 1.0D+00
      one24 = one/24.0D+00
      sum = 0.0D+00
      do 10 i = 1,n-1
         diff = xa(i+1)-xa(i)
         sum = sum+diff*( (ya(i)+ya(i+1))/2.0D+00 -
     &         one24*diff*diff*(y2a(i)+y2a(i+1)))
 10   continue

      xint = sum

      return
      end

      subroutine calcsd(x,y,ydp,a,b,c,r,n,an,c1)
      implicit real*8 (a-h,o-z)
      dimension x(n),y(n),ydp(n),a(n),b(n),c(n),r(n)

      a(1) = 0.0D+00
      a(n) = an
      b(1) = 1.0D+00
      b(n) = 1.0D+00
      c(1) = c1
      c(n) = 0.0D+00
      r(1) = 0.0D+00
      r(n) = 0.0D+00

      do 21 i = 2,n-1
         a(i) = (x(i)-x(i-1))/6.0D+00
         b(i) = (x(i+1)-x(i-1))/3.0D+00
         c(i) = (x(i+1)-x(i))/6.0D+00
         r(i) = (y(i+1)-y(i))/(x(i+1)-x(i))-
     &          (y(i)-y(i-1))/(x(i)-x(i-1))
 21   continue

      call tridag(a,b,c,r,ydp,n)

      return
      end

      subroutine spling(xa,ya,y2a,x,y,yp,ydp,khi,klo,kvals,n,
     &                  ndoy,ndoyp,ndoydp)
      implicit real*8 (a-h,o-z)
      dimension xa(n),ya(n),y2a(n)

      one = 1.0D+00
      if (kvals .eq. 0) then
         klo = 1
         khi = n
 1       if ((khi-klo) .gt. 1) then
            k = (khi+klo)/2
            if (xa(k) .gt. x) then
               khi = k
            else
               klo = k
            endif
            goto 1
         endif
      endif

      h = xa(khi)-xa(klo)
      a = (xa(khi)-x)/h
      b = (x-xa(klo))/h
      asq = a*a
      bsq = b*b

      if (ndoy .eq. 1)
     &   y = a*ya(klo)+b*ya(khi)+
     &       ((asq*a-a)*y2a(klo)+
     &       (bsq*b-b)*y2a(khi))*(h*h)/6.0D+00
      if (ndoyp .eq. 1)
     &   yp = (ya(khi)-ya(klo))/h-
     &        (3.0D+00*asq-one)/6.0D+00*h*y2a(klo)+
     &        (3.0D+00*bsq-one)/6.0D+00*h*y2a(khi)
      if (ndoydp .eq. 1)
     &   ydp = a*y2a(klo)+b*y2a(khi)
      
      return
      end

	double precision function gmm(theta)
	implicit real*8 (a-h,o-z)
	parameter (ncapt=10000,ndrop=500,one=1.0D+00)
	common/nshock/nz(ncapt+1)
	common/cgmm/rent(ncapt),cons(ncapt)
	common/vgmm/vhat(3,3),wght(3,3)
      dimension theta(2),fval(ncapt,3)

	beta = theta(1)
	gamma = theta(2)

      npts = 0
	do 10 i = ndrop+1,ncapt-1
	   npts = npts+1
	   fval(npts,1) = one-beta*(cons(i+1)/cons(i))**(-gamma)*
     &                           (one+rent(i+1))
	   fval(npts,2) = fval(npts,1)*rent(i)
	   fval(npts,3) = fval(npts,1)*cons(i)/cons(i-1)
 10   continue

      gmm = 0.0D+00
      do 11 i = 1,npts
	   gmm = gmm+fval(i,1)*wght(1,1)*fval(i,1)+
     &             fval(i,1)*wght(1,2)*fval(i,2)+
     &	         fval(i,1)*wght(1,3)*fval(i,3)+
     &             fval(i,2)*wght(2,1)*fval(i,1)+
     &             fval(i,2)*wght(2,2)*fval(i,2)+
     &             fval(i,2)*wght(2,3)*fval(i,3)+
     &             fval(i,3)*wght(3,1)*fval(i,1)+
     &             fval(i,3)*wght(3,2)*fval(i,2)+
     &             fval(i,3)*wght(3,3)*fval(i,3)
 11   continue

	return
	end

c     Simulation for particular sequence

      subroutine calcf2(optkg1,optkg0,xk2pts,xm2pts,
     &                  a0,b0,xkmean,sdmean,
     &				  skmean,s4mean,s5mean,rsqa,rsqb,sderra,sderrb,
     &                  gmean,p1mean,p5mean,p10mean,p20mean,p30mean,
     &                  p90mean,p95mean,hansen,erfmx1,erfmx2)
      implicit real*8 (a-h,o-z)
      parameter (nk2pts=5000,ncapt=10000,ndrop=500,nm2pts=30,
     &           unempb=0.1D+00,unempg=0.04D+00,durub=2.5D+00,
     &           durug=1.5D+00,nread2=1,alpha=0.36D+00,xmu=0.15D+00,
     &           hfix=1.0D+00/0.9D+00,delta=0.025D+00,zgood=1.01D+00,
     &           zbad=0.99D+00,gamma=1.0D+00,beta=0.99D+00)
      common/cfs/f1(nk2pts,2)
      common/cprob/pi(4,4)
	common/nshock/nz(ncapt+1),nind(ncapt)
	common/cvgrid/vgrid(nk2pts,nm2pts,4)
	common/cgmm/rent(ncapt),cons(ncapt)
	common/vgmm/vhat(3,3),wght(3,3)
	external gmm
      dimension a0(2),b0(2),
     &          optkg1(nk2pts,nm2pts,2),abdev(ncapt),
     &          optkg0(nk2pts,nm2pts,2),fracs(ncapt),
     &          xk2pts(nk2pts),xm2pts(nm2pts),
     &          xkvals(ncapt+1),f0(nk2pts,2),stdxk(ncapt),
     &          skewxk(ncapt),xkurtk(ncapt),ginis(ncapt),
     &          pct1(ncapt),pct5(ncapt),pct10(ncapt),pct20(ncapt),
     &          pct30(ncapt),toppct(7),tops(7),
     &          vmus(ncapt),xkbare(ncapt),xkbaru(ncapt),
     &          xkind(ncapt+1),cind(ncapt),xkfive(ncapt),pct90(ncapt),
     &          pct95(ncapt),xkpred(ncapt+1)

      one = 1.0D+00
	xkinc = xk2pts(2)-xk2pts(1)
      xminc = xm2pts(2)-xm2pts(1)
	toppct(1) = 0.01D+00
	toppct(2) = 0.05D+00
	toppct(3) = 0.1D+00
	toppct(4) = 0.2D+00
	toppct(5) = 0.3D+00
	toppct(6) = 0.9D+00
	toppct(7) = 0.95D+00

      pgg00 = (durug-one)/durug
	pbb00 = (durub-one)/durub
	pbg00 = 1.25D+00*pbb00
	pgb00 = 0.75D+00*pgg00
	pgg01 = (unempg-unempg*pgg00)/(one-unempg)
	pbb01 = (unempb-unempb*pbb00)/(one-unempb)
	pbg01 = (unempb-unempg*pbg00)/(one-unempg)
	pgb01 = (unempg-unempb*pgb00)/(one-unempb)

	pgg10 = one-(durug-one)/durug
	pbb10 = one-(durub-one)/durub
	pbg10 = one-1.25D+00*pbb00
	pgb10 = one-0.75D+00*pgg00
	pgg11 = one-(unempg-unempg*pgg00)/(one-unempg)
	pbb11 = one-(unempb-unempb*pbb00)/(one-unempb)
	pbg11 = one-(unempb-unempg*pbg00)/(one-unempg)
	pgb11 = one-(unempg-unempb*pgb00)/(one-unempb)

      xkmean = 0.0D+00
	sdmean = 0.0D+00
	skmean = 0.0D+00
	s4mean = 0.0D+00
	s5mean = 0.0D+00
	gmean = 0.0D+00
	p1mean = 0.0D+00
	p5mean = 0.0D+00
	p10mean = 0.0D+00
	p20mean = 0.0D+00
	p30mean = 0.0D+00
	p90mean = 0.0D+00
	p95mean = 0.0D+00
	
	contot = 0.0D+00
	congd = 0.0D+00
	conbd = 0.0D+00
	mgood = 0
	mbad = 0

	xkmne = 0.0D+00
	xkmnu = 0.0D+00

	xk5em = 0.0D+00
	xk10em = 0.0D+00
	xk5um = 0.0D+00
	xk10um = 0.0D+00

      open(2,file='initialdist.f1',form='formatted')
      do 66 i = 1,nk2pts
         read(2,547) dum,f1(i,2),f1(i,1)
 547     format(3f18.10)
 66   continue
      close(2)
	fsum = 0.0D+00
	do 38 i = 1,nk2pts
          fsum = fsum+f1(i,1)+f1(i,2)
 38   continue
      if (dabs(fsum-one) .gt. 1.0D-04) then
	   write(6,575) fsum
 575     format('initial distribution integrates to :',f18.10)
         pause
	endif

      xkbcur = 0.0D+00
	do 45 i = 1,nk2pts
	   xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
 45   continue

      xkind(1) = 43.0D+00
	xkpred(1) = xkbcur

      open(7,file='jedc_final_1.sim22',form='formatted')
	open(8,file='jedc_final_1.ind22',form='formatted')
	open(9,file='jedc_final_1.pec22',form='formatted')
      do 12 ii = 1,ncapt

         nzcur = nz(ii)
         if (nzcur .eq. 1) then
	      nz2 = 1
	      nz3 = 2
	   else
	      nz2 = 3
	      nz3 = 4
	   endif
	   if (nzcur .eq. 1) then   
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgg11
	         prob1 = pgg10
	      else
	         prob0 = pbg11
	         prob1 = pbg10
	      endif
	   else
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgb11
	         prob1 = pgb10
	      else
	         prob0 = pbb11
	         prob1 = pbb10
	      endif
         endif
         xkvals(ii) = xkbcur

         if (nz(ii) .eq. 1) then
	      rent(ii) = alpha*zgood*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempg))**(one-alpha)-delta
	      wage = (one-alpha)*zgood*xkvals(ii)**alpha*
     &             (hfix*(one-unempg))**(-alpha)
	      gdp = zgood*xkvals(ii)**alpha*(hfix*(one-unempg))**
     &            (one-alpha)
	      tau = xmu*unempg/((one-unempg)*hfix)
	   else
	      rent(ii) = alpha*zbad*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempb))**(one-alpha)-delta
	      wage = (one-alpha)*zbad*xkvals(ii)**alpha*
     &             (hfix*(one-unempb))**(-alpha)
	      gdp = zbad*xkvals(ii)**alpha*(hfix*(one-unempb))**
     &            (one-alpha)
	      tau = xmu*unempb/((one-unempb)*hfix)
	   endif

         f0 = f1
         f1 = 0.0D+00

c        Determine location of current K in K grid

         call hunt(xm2pts,nm2pts,xkbcur,mcor)
         if (mcor .lt. 1) then
            capa = 0.0D+00
            mcor = 1
         elseif (mcor .ge. nm2pts) then
            capa = 1.0D+00
            mcor = nm2pts-1
         else
            capa = (xkbcur-xm2pts(mcor))/xminc
         endif

c        Use decision rules to update distribution

         vmu = 0.0D+00
         do 88 i = 1,nk2pts

            xk1 = optkg1(i,mcor,nzcur)+capa*
     &            (optkg1(i,mcor+1,nzcur)-optkg1(i,mcor,nzcur))
            xk0 = optkg0(i,mcor,1)+capa*
     &            (optkg0(i,mcor+1,nzcur)-optkg0(i,mcor,nzcur))
	      v1 = vgrid(i,mcor,nz2)+capa*
     &           (vgrid(i,mcor+1,nz2)-vgrid(i,mcor,nz2))
	      v0 = vgrid(i,mcor,nz3)+capa*
     &           (vgrid(i,mcor+1,nz3)-vgrid(i,mcor,nz3))
            
            call hunt(xk2pts,nk2pts,xk1,jlo1)
            if (jlo1 .lt. 1) then
               wght1 = 1.0D+00
               jlo1 = 1
            elseif (jlo1 .ge. nk2pts) then
               wght1 = 0.0D+00
               jlo1 = nk2pts-1
            else
               wght1 = one-(xk1-xk2pts(jlo1))/xkinc
            endif

            call hunt(xk2pts,nk2pts,xk0,jlo0)
            if (jlo0 .lt. 1) then
               wght0 = 1.0D+00
               jlo0 = 1
            elseif (jlo0 .ge. nk2pts) then
               wght0 = 0.0D+00
               jlo0 = nk2pts-1
            else
               wght0 = one-(xk0-xk2pts(jlo0))/xkinc
            endif

            f1(jlo1,1) = prob0*wght1*f0(i,1)+f1(jlo1,1)
            f1(jlo1+1,1) = prob0*(one-wght1)*f0(i,1)+
     &                     f1(jlo1+1,1)
            f1(jlo0,1) = prob1*wght0*f0(i,2)+f1(jlo0,1)
            f1(jlo0+1,1) = prob1*(one-wght0)*f0(i,2)+
     &                     f1(jlo0+1,1)
            
		  f1(jlo1,2) = (one-prob0)*wght1*f0(i,1)+f1(jlo1,2)
            f1(jlo1+1,2) = (one-prob0)*(one-wght1)*
     &                      f0(i,1)+f1(jlo1+1,2)
            f1(jlo0,2) = (one-prob1)*wght0*f0(i,2)+f1(jlo0,2)
            f1(jlo0+1,2) = (one-prob1)*(one-wght0)*
     &                      f0(i,2)+f1(jlo0+1,2)

            vmu = vmu+v1*f0(i,1)+v0*f0(i,2)

 88      continue
         vmus(ii) = vmu

c        Compute individual behavior
         call hunt(xk2pts,nk2pts,xkind(ii),jk)
	   if (jk .lt. 1) then
	      jk = 1
	      capk = 0.0D+00
	   elseif (jk .ge. nk2pts) then
	      jk = nk2pts-1
	      capk = 1.0D+00
	   else
	      capk = (xkind(ii)-xk2pts(jk))/xkinc
	   endif
	   if (nind(ii) .eq. 1) then
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg1(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg1(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg1(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg1(jk+1,mcor+1,nzcur)
	      yind = (one+rent(ii))*xkind(ii)+(one-tau)*wage*hfix
	      cind(ii) = yind-xkind(ii+1)
	   else
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg0(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg0(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg0(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg0(jk+1,mcor+1,nzcur)
	      yind = (one+rent(ii))*xkind(ii)+xmu*wage
	      cind(ii) = yind-xkind(ii+1)
	   endif

c        Compute K' from new distribution f1

         xkbcur = 0.0D+00
	   fsum = 0.0D+00
         do 83 i = 1,nk2pts
            xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
	      fsum = fsum+f1(i,1)+f1(i,2)
 83      continue
         if (dabs(fsum-one) .gt. 1.0D-04) then
	      write(6,783) ii,fsum
 783        format('does not integrate to one in period ',i6,f18.10)
            pause
	   endif
	   if (nz(ii) .eq. 1) then
	      xkpred(ii+1) = dexp(a0(1)+a0(2)*dlog(xkpred(ii)))
	   else
	      xkpred(ii+1) = dexp(b0(1)+b0(2)*dlog(xkpred(ii)))
	   endif
         xinv = xkbcur-(one-delta)*xkvals(ii)
	   cons(ii) = gdp-xinv

         varcur = 0.0D+00
         skwcur = 0.0D+00
         xkurt = 0.0D+00
	   unemp = 0.0D+00
	   frpoor = 0.0D+00
	   absval = 0.0D+00
	   fiver = 0.0D+00
	   xke = 0.0D+00
	   xku = 0.0D+00
         do 82 i = 1,nk2pts
	      fterm = f0(i,1)+f0(i,2)
	      xterm = xk2pts(i)-xkvals(ii)
            varcur = varcur+fterm*xterm*xterm
            skwcur = skwcur+fterm*xterm*xterm*xterm
            xkurt = xkurt+fterm*xterm*xterm*xterm*xterm
	      fiver = fiver+fterm*xterm*xterm*xterm*xterm*xterm
	      unemp = unemp+f0(i,2)
	      absval = absval+dabs(xterm)*fterm
	      if (xk2pts(i) .lt. 0.1D+00*xkvals(ii)) frpoor = frpoor+fterm
	      xke = xke+xk2pts(i)*f0(i,1)
	      xku = xku+xk2pts(i)*f0(i,2)
 82      continue
         stdxk(ii) = dsqrt(varcur)/xkvals(ii)
         skewxk(ii) = skwcur**(one/3.0D+00)/xkvals(ii)
         xkurtk(ii) = xkurt**(one/4.0D+00)/xkvals(ii)
	   xkfive(ii) = fiver**(one/5.0D+00)/xkvals(ii)
	   fracs(ii) = frpoor
	   abdev(ii) = absval
	   xkbare(ii) = xke/(one-unemp)
	   xkbaru(ii) = xku/unemp

         call getgini(f0,xk2pts,xkvals(ii),ginis(ii),toppct,tops)
         pct1(ii) = tops(1)
         pct5(ii) = tops(2)
	   pct10(ii) = tops(3)
	   pct20(ii) = tops(4)
	   pct30(ii) = tops(5)
	   pct90(ii) = tops(6)
	   pct95(ii) = tops(7)

         if (nz(ii) .eq. 1) then
	      nnnz = 2
	   else
	      nnnz = 1
	   endif
	   if (nind(ii) .eq. 1) then
	      nnni = 2
	   else
	      nnni = 1
	   endif

	   write(7,504) ii,nz(ii),xkvals(ii),
     &                stdxk(ii),skewxk(ii),xkurtk(ii),xkfive(ii),unemp,
     &                fracs(ii),abdev(ii),rent(ii),wage,ginis(ii),gdp,
     &                cons(ii),xinv,pct1(ii),pct5(ii),pct10(ii),
     &                pct20(ii),pct30(ii),pct90(ii),pct95(ii),vmus(ii),
     &                xkpred(ii),xkbare(ii),xkbaru(ii)
 504     format(i6,i3,25f15.9)
         write(8,556) ii,nz(ii),nind(ii),xkind(ii),cind(ii),yind,
     &                cons(ii),gdp,xkvals(ii)
 556     format(i6,2i3,6f15.9)
         write(9,399) nnnz,xkvals(ii),xkpred(ii)
 399     format(i3,2f18.10)

 12   continue
      close(7)
	close(8)
	close(9)

      err1mx = 0.0D+00
	err2mx = 0.0D+00
	err1sm = 0.0D+00
	err2sm = 0.0D+00
	do 190 i = 1,200
	   abserr = dabs(xkvals(i)-xkpred(i))
	   relerr = dabs((xkvals(i)-xkpred(i))/xkvals(i))
	   if (abserr .gt. err1mx) err1mx = abserr
	   if (relerr .gt. err2mx) err2mx = relerr
	   err1sm = err1sm+abserr/dble(real(200))
	   err2sm = err2sm+relerr/dble(real(200))
 190  continue
      write(6,253) err1mx,err2mx
 253  format('max errors, particular: ',2f18.10)
      write(6,252) err1sm,err2sm
 252  format('avg errors, particular: ',2f18.10)

      return
      end

c     Simulation for random sequence

      subroutine calcf3(optkg1,optkg0,xk2pts,xm2pts,a0,b0,
     &                  xkmean,sdmean,skmean,s4mean,s5mean,rsqa,rsqb,
     &                  sderra,sderrb,
     &                  gmean,p1mean,p5mean,p10mean,p20mean,
     &                  p30mean,p90mean,p95mean,hansen,erfmx1,erfmx2)
      implicit real*8 (a-h,o-z)
      parameter (nk2pts=5000,ncapt=10000,ndrop=500,nm2pts=30,
     &           unempb=0.1D+00,unempg=0.04D+00,durub=2.5D+00,
     &           durug=1.5D+00,nread2=1,alpha=0.36D+00,xmu=0.15D+00,
     &           hfix=1.0D+00/0.9D+00,delta=0.025D+00,zgood=1.01D+00,
     &           zbad=0.99D+00,gamma=1.0D+00,beta=0.99D+00)
      common/cfs/f1(nk2pts,2)
      common/cprob/pi(4,4)
	common/nshock/nz(ncapt+1),nind(ncapt)
	common/cvgrid/vgrid(nk2pts,nm2pts,4)
	common/cgmm/rent(ncapt),cons(ncapt)
	common/vgmm/vhat(3,3),wght(3,3)
	external gmm
      dimension a0(2),b0(2),a0new(2),b0new(2),atinv(28,28),
     &          optkg1(nk2pts,nm2pts,2),abdev(ncapt),temp(1,28),
     &          optkg0(nk2pts,nm2pts,2),fracs(ncapt),btt(1,28),
     &          xk2pts(nk2pts),xm2pts(nm2pts),xdata3(ncapt,28),
     &          xkvals(ncapt+1),f0(nk2pts,2),xdata2(ncapt,28),
     &          xdata(ncapt,2),ydata(ncapt),stdxk(ncapt),ttt(1,1),
     &          skewxk(ncapt),xkurtk(ncapt),ginis(ncapt),
     &          pct1(ncapt),pct5(ncapt),pct10(ncapt),pct20(ncapt),
     &          pct30(ncapt),toppct(7),tops(7),bt(28,1),at(28,28),
     &          vmus(ncapt),xdatan(ncapt,6),ydatan(ncapt),bhat(6),
     &          xdatas(ncapt,8),p0new(8),q0new(8),p(3,2),y(3),
     &          xkind(ncapt+1),cind(ncapt),xkfive(ncapt),pct90(ncapt),
     &          pct95(ncapt),xkpred(ncapt+1),xkbare(ncapt),
     &          xkbaru(ncapt),wage(ncapt),gdp(ncapt),rpred(ncapt),
     &          yind(ncapt),cgrow(ncapt),xinv(ncapt),wpred(ncapt),
     &          ypred(ncapt),cpred(ncapt),xpred(ncapt),cumfe(nk2pts),
     &          cumfu(nk2pts),xk5e(ncapt),xk10e(ncapt),xk5u(ncapt),
     &          xk10u(ncapt),cumf(nk2pts)
	dimension rhocici(3),rhokiki(3),rhoym(3)
      dimension stdxke(ncapt),stdxku(ncapt),skwxke(ncapt),
     &          skwxku(ncapt),xkwxke(ncapt),xkwxku(ncapt),
     &          x5wxke(ncapt),x5wxku(ncapt),xk5(ncapt),xk10(ncapt)

      write(6,392)
 392  format('beginning random simulation')

      one = 1.0D+00
	xkinc = xk2pts(2)-xk2pts(1)
      xminc = xm2pts(2)-xm2pts(1)
	toppct(1) = 0.01D+00
	toppct(2) = 0.05D+00
	toppct(3) = 0.1D+00
	toppct(4) = 0.2D+00
	toppct(5) = 0.3D+00
	toppct(6) = 0.9D+00
	toppct(7) = 0.95D+00

c     Individual state transition probabilities

      pgg00 = (durug-one)/durug
	pbb00 = (durub-one)/durub
	pbg00 = 1.25D+00*pbb00
	pgb00 = 0.75D+00*pgg00
	pgg01 = (unempg-unempg*pgg00)/(one-unempg)
	pbb01 = (unempb-unempb*pbb00)/(one-unempb)
	pbg01 = (unempb-unempg*pbg00)/(one-unempg)
	pgb01 = (unempg-unempb*pgb00)/(one-unempb)

	pgg10 = one-(durug-one)/durug
	pbb10 = one-(durub-one)/durub
	pbg10 = one-1.25D+00*pbb00
	pgb10 = one-0.75D+00*pgg00
	pgg11 = one-(unempg-unempg*pgg00)/(one-unempg)
	pbb11 = one-(unempb-unempb*pbb00)/(one-unempb)
	pbg11 = one-(unempb-unempg*pbg00)/(one-unempg)
	pgb11 = one-(unempg-unempb*pgb00)/(one-unempb)

      xkmean = 0.0D+00
	sdmean = 0.0D+00
	skmean = 0.0D+00
	s4mean = 0.0D+00
	s5mean = 0.0D+00
	gmean = 0.0D+00
	p1mean = 0.0D+00
	p5mean = 0.0D+00
	p10mean = 0.0D+00
	p20mean = 0.0D+00
	p30mean = 0.0D+00
	p90mean = 0.0D+00
	p95mean = 0.0D+00

	emean1 = 0.0D+00
	emean2 = 0.0D+00
	emean3 = 0.0D+00
	emean4 = 0.0D+00
	emean5 = 0.0D+00
	umean1 = 0.0D+00
	umean2 = 0.0D+00
	umean3 = 0.0D+00
	umean4 = 0.0D+00
	umean5 = 0.0D+00
	
	contot = 0.0D+00
	congd = 0.0D+00
	conbd = 0.0D+00
	mgood = 0
	mbad = 0

	xkmne = 0.0D+00
	xkmnu = 0.0D+00

      xk5m = 0.0D+00
	xk10m = 0.0D+00
	xk5em = 0.0D+00
	xk10em = 0.0D+00
	xk5um = 0.0D+00
	xk10um = 0.0D+00

	xk5eg = 0.0D+00
	xk10eg = 0.0D+00
	xk5ug = 0.0D+00
	xk10ug = 0.0D+00
	xk5g = 0.0D+00
	xk10g = 0.0D+00
	xk5eb = 0.0D+00
	xk10eb = 0.0D+00
	xk5ub = 0.0D+00
	xk10ub = 0.0D+00
	xk5b = 0.0D+00
	xk10b = 0.0D+00

c     Initial distribution

      open(2,file='initialdist.f1',form='formatted')
      do 66 i = 1,nk2pts
         read(2,547) dum,f1(i,2),f1(i,1)
 547     format(3f18.10)
 66   continue
      close(2)
	fsum = 0.0D+00
	do 38 i = 1,nk2pts
          fsum = fsum+f1(i,1)+f1(i,2)
 38   continue
      if (dabs(fsum-one) .gt. 1.0D-04) then
	   write(6,575) fsum
 575     format('initial distribution integrates to :',f18.10)
         pause
	endif

      xkbcur = 0.0D+00
	do 45 i = 1,nk2pts
	   xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
 45   continue

c     Simulation loop
      xkind(1) = 43.0D+00
	xkpred(1) = xkbcur

      open(7,file='jedc_final_1.sim33',form='formatted')
	open(8,file='jedc_final_1.ind33',form='formatted')
	open(9,file='jedc_final_1.wkm33',form='formatted')
	open(10,file='jedc_final_1.ran33',form='formatted')
	open(1,file='jedc_final_1.ucc33',form='formatted')
      do 12 ii = 1,ncapt

         nzcur = nz(ii)
         if (nzcur .eq. 1) then
	      nz2 = 1
	      nz3 = 2
	   else
	      nz2 = 3
	      nz3 = 4
	   endif
	   if (nzcur .eq. 1) then   
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgg11
	         prob1 = pgg10
	      else
	         prob0 = pbg11
	         prob1 = pbg10
	      endif
	   else
            if (nz(ii+1) .eq. 1) then
	         prob0 = pgb11
	         prob1 = pgb10
	      else
	         prob0 = pbb11
	         prob1 = pbb10
	      endif
         endif
         xkvals(ii) = xkbcur

         if (nz(ii) .eq. 1) then
	      rent(ii) = alpha*zgood*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempg))**(one-alpha)-delta
	      wage(ii) = (one-alpha)*zgood*xkvals(ii)**alpha*
     &                 (hfix*(one-unempg))**(-alpha)
	      gdp(ii) = zgood*xkvals(ii)**alpha*(hfix*(one-unempg))**
     &                (one-alpha)
	      tau = xmu*unempg/((one-unempg)*hfix)
	   else
	      rent(ii) = alpha*zbad*xkvals(ii)**(alpha-one)*
     &                 (hfix*(one-unempb))**(one-alpha)-delta
	      wage(ii) = (one-alpha)*zbad*xkvals(ii)**alpha*
     &                 (hfix*(one-unempb))**(-alpha)
	      gdp(ii) = zbad*xkvals(ii)**alpha*(hfix*(one-unempb))**
     &                (one-alpha)
	      tau = xmu*unempb/((one-unempb)*hfix)
	   endif

         f0 = f1
         f1 = 0.0D+00

c        Determine location of current K in K grid

         call hunt(xm2pts,nm2pts,xkbcur,mcor)
         if (mcor .lt. 1) then
            capa = 0.0D+00
            mcor = 1
         elseif (mcor .ge. nm2pts) then
            capa = 1.0D+00
            mcor = nm2pts-1
         else
            capa = (xkbcur-xm2pts(mcor))/xminc
         endif

c        Use decision rules to update distribution

         vmu = 0.0D+00
         do 88 i = 1,nk2pts

            xk1 = optkg1(i,mcor,nzcur)+capa*
     &            (optkg1(i,mcor+1,nzcur)-optkg1(i,mcor,nzcur))
            xk0 = optkg0(i,mcor,1)+capa*
     &            (optkg0(i,mcor+1,nzcur)-optkg0(i,mcor,nzcur))
	      v1 = vgrid(i,mcor,nz2)+capa*
     &           (vgrid(i,mcor+1,nz2)-vgrid(i,mcor,nz2))
	      v0 = vgrid(i,mcor,nz3)+capa*
     &           (vgrid(i,mcor+1,nz3)-vgrid(i,mcor,nz3))
            
            call hunt(xk2pts,nk2pts,xk1,jlo1)
            if (jlo1 .lt. 1) then
               wght1 = 1.0D+00
               jlo1 = 1
            elseif (jlo1 .ge. nk2pts) then
               wght1 = 0.0D+00
               jlo1 = nk2pts-1
            else
               wght1 = one-(xk1-xk2pts(jlo1))/xkinc
            endif

            call hunt(xk2pts,nk2pts,xk0,jlo0)
            if (jlo0 .lt. 1) then
               wght0 = 1.0D+00
               jlo0 = 1
            elseif (jlo0 .ge. nk2pts) then
               wght0 = 0.0D+00
               jlo0 = nk2pts-1
            else
               wght0 = one-(xk0-xk2pts(jlo0))/xkinc
            endif

            f1(jlo1,1) = prob0*wght1*f0(i,1)+f1(jlo1,1)
            f1(jlo1+1,1) = prob0*(one-wght1)*f0(i,1)+
     &                     f1(jlo1+1,1)
            f1(jlo0,1) = prob1*wght0*f0(i,2)+f1(jlo0,1)
            f1(jlo0+1,1) = prob1*(one-wght0)*f0(i,2)+
     &                     f1(jlo0+1,1)
            
		  f1(jlo1,2) = (one-prob0)*wght1*f0(i,1)+f1(jlo1,2)
            f1(jlo1+1,2) = (one-prob0)*(one-wght1)*
     &                      f0(i,1)+f1(jlo1+1,2)
            f1(jlo0,2) = (one-prob1)*wght0*f0(i,2)+f1(jlo0,2)
            f1(jlo0+1,2) = (one-prob1)*(one-wght0)*
     &                      f0(i,2)+f1(jlo0+1,2)

            vmu = vmu+v1*f0(i,1)+v0*f0(i,2)

 88      continue
         vmus(ii) = vmu

c        Compute individual behavior
         call hunt(xk2pts,nk2pts,xkind(ii),jk)
	   if (jk .lt. 1) then
	      jk = 1
	      capk = 0.0D+00
	   elseif (jk .ge. nk2pts) then
	      jk = nk2pts-1
	      capk = 1.0D+00
	   else
	      capk = (xkind(ii)-xk2pts(jk))/xkinc
	   endif
	   if (nind(ii) .eq. 1) then
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg1(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg1(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg1(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg1(jk+1,mcor+1,nzcur)
	      yind(ii) = (one+rent(ii))*xkind(ii)+(one-tau)*wage(ii)*hfix
	      cind(ii) = yind(ii)-xkind(ii+1)
	   else
	      xkind(ii+1) = (one-capk)*(one-capa)*optkg0(jk,mcor,nzcur)+
     &                    capk*(one-capa)*optkg0(jk+1,mcor,nzcur)+
     &                    (one-capk)*capa*optkg0(jk,mcor+1,nzcur)+
     &                    capk*capa*optkg0(jk+1,mcor+1,nzcur)
	      yind(ii) = (one+rent(ii))*xkind(ii)+xmu*wage(ii)
	      cind(ii) = yind(ii)-xkind(ii+1)
	   endif

c        Compute K' from new distribution f1

         xkbcur = 0.0D+00
	   fsum = 0.0D+00
         do 83 i = 1,nk2pts
            xkbcur = xkbcur+xk2pts(i)*(f1(i,1)+f1(i,2))
	      fsum = fsum+f1(i,1)+f1(i,2)
 83      continue
         if (dabs(fsum-one) .gt. 1.0D-04) then
	      write(6,783) ii,fsum
 783        format('does not integrate to one in period ',i6,f18.10)
            pause
	   endif
	   if (nz(ii) .eq. 1) then
	      xkpred(ii+1) = dexp(a0(1)+a0(2)*dlog(xkpred(ii)))
	   else
	      xkpred(ii+1) = dexp(b0(1)+b0(2)*dlog(xkpred(ii)))
	   endif
         xinv(ii) = xkbcur-(one-delta)*xkvals(ii)
	   cons(ii) = gdp(ii)-xinv(ii)

C        Centered and normalized aggregate moments
C        Also conditional means

         varcur = 0.0D+00
         skwcur = 0.0D+00
         xkurt = 0.0D+00
	   unemp = 0.0D+00
	   frpoor = 0.0D+00
	   absval = 0.0D+00
	   fiver = 0.0D+00
	   xke = 0.0D+00
	   xku = 0.0D+00
         do 82 i = 1,nk2pts
	      fterm = f0(i,1)+f0(i,2)
	      xterm = xk2pts(i)-xkvals(ii)
            varcur = varcur+fterm*xterm*xterm
            skwcur = skwcur+fterm*xterm*xterm*xterm
            xkurt = xkurt+fterm*xterm*xterm*xterm*xterm
	      fiver = fiver+fterm*xterm*xterm*xterm*xterm*xterm
	      unemp = unemp+f0(i,2)
	      absval = absval+dabs(xterm)*fterm
	      if (xk2pts(i) .lt. 0.1D+00*xkvals(ii)) frpoor = frpoor+fterm
	      xke = xke+xk2pts(i)*f0(i,1)
	      xku = xku+xk2pts(i)*f0(i,2)
 82      continue
         stdxk(ii) = dsqrt(varcur)/xkvals(ii)
         skewxk(ii) = skwcur**(one/3.0D+00)/xkvals(ii)
         xkurtk(ii) = xkurt**(one/4.0D+00)/xkvals(ii)
	   xkfive(ii) = fiver**(one/5.0D+00)/xkvals(ii)
	   fracs(ii) = frpoor
	   abdev(ii) = absval
	   xkbare(ii) = xke/(one-unemp)
	   xkbaru(ii) = xku/unemp

c        Centered and normalized conditional moments

	   vke = 0.0D+00
	   vku = 0.0D+00
	   ske = 0.0D+00
	   sku = 0.0D+00
	   xkke = 0.0D+00
	   xkku = 0.0D+00
	   x5e = 0.0D+00
	   x5u = 0.0D+00
         do 813 i = 1,nk2pts
	      xterme = xk2pts(i)-xkbare(ii)
	      xtermu = xk2pts(i)-xkbaru(ii)
	      vke = vke+xterme*xterme*f0(i,1)
	      vku = vku+xtermu*xtermu*f0(i,2)
	      ske = ske+xterme*xterme*xterme*f0(i,1)
	      sku = sku+xtermu*xtermu*xtermu*f0(i,2)
            xkke = xkke+xterme*xterme*xterme*xterme*f0(i,1)
	      xkku = xkku+xtermu*xtermu*xtermu*xtermu*f0(i,2)
            x5e = x5e+xterme*xterme*xterme*xterme*xterme*f0(i,1)
	      x5u = x5u+xtermu*xtermu*xtermu*xtermu*xtermu*f0(i,2)
 813     continue
         stdxke(ii) = dsqrt(vke/(one-unemp))/xkbare(ii)
	   stdxku(ii) = dsqrt(vku/unemp)/xkbaru(ii)
	   skwxke(ii) = (ske/(one-unemp))**(one/3.0D+00)/xkbare(ii)
	   skwxku(ii) = (sku/unemp)**(one/3.0D+00)/xkbaru(ii)
	   xkwxke(ii) = (xkke/(one-unemp))**(one/4.0D+00)/xkbare(ii)
	   xkwxku(ii) = (xkku/unemp)**(one/4.0D+00)/xkbaru(ii)
         x5wxke(ii) = (x5e/(one-unemp))**(one/5.0D+00)/xkbare(ii)
	   x5wxku(ii) = (x5u/unemp)**(one/5.0D+00)/xkbaru(ii)

C        Uncentered and unnormalized conditional moments
         
         xk2euc = 0.0D+00
	   xk2uuc = 0.0D+00
	   xk3euc = 0.0D+00
	   xk3uuc = 0.0D+00
	   xk4euc = 0.0D+00
	   xk4uuc = 0.0D+00
	   xk5euc = 0.0D+00
	   xk5uuc = 0.0D+00
	   do 918 i = 1,nk2pts
	      xk2euc = xk2euc+xk2pts(i)*xk2pts(i)*f0(i,1)/(one-unemp)
	      xk2uuc = xk2uuc+xk2pts(i)*xk2pts(i)*f0(i,2)/unemp
	      xk3euc = xk3euc+xk2pts(i)*xk2pts(i)*xk2pts(i)*f0(i,1)/
     &               (one-unemp)
	      xk3uuc = xk3uuc+xk2pts(i)*xk2pts(i)*xk2pts(i)*f0(i,2)/unemp
	      xk4euc = xk4euc+xk2pts(i)*xk2pts(i)*xk2pts(i)*
     &               xk2pts(i)*f0(i,1)/(one-unemp)
	      xk4uuc = xk4uuc+xk2pts(i)*xk2pts(i)*xk2pts(i)*
     &               xk2pts(i)*f0(i,2)/unemp
	      xk5euc = xk5euc+xk2pts(i)*xk2pts(i)*xk2pts(i)*
     &               xk2pts(i)*xk2pts(i)*f0(i,1)/(one-unemp)
	      xk5uuc = xk5uuc+xk2pts(i)*xk2pts(i)*xk2pts(i)*
     &               xk2pts(i)*xk2pts(i)*f0(i,2)/unemp
 918     continue

C        Gini coefficient and Lorenz curve

	   call getgini(f0,xk2pts,xkvals(ii),ginis(ii),toppct,tops)
         pct1(ii) = tops(1)
         pct5(ii) = tops(2)
	   pct10(ii) = tops(3)
	   pct20(ii) = tops(4)
	   pct30(ii) = tops(5)
	   pct90(ii) = tops(6)
	   pct95(ii) = tops(7)

C        Percentiles

         cumfe = 0.0D+00
	   cumfu = 0.0D+00
	   cumfe(1) = f0(1,1)/(one-unemp)
	   cumfu(1) = f0(1,2)/unemp
         do 94 i = 2,nk2pts
	      cumfe(i) = cumfe(i-1)+f0(i,1)/(one-unemp)
	      cumfu(i) = cumfu(i-1)+f0(i,2)/unemp
 94      continue
         call hunt(cumfe,nk2pts,0.05D+00,je5)
	   ewgt5 = (0.05D+00-cumfe(je5))/(cumfe(je5+1)-cumfe(je5))
	   call hunt(cumfe,nk2pts,0.1D+00,je10)
	   ewgt10 = (0.1D+00-cumfe(je10))/(cumfe(je10+1)-cumfe(je10))
         call hunt(cumfu,nk2pts,0.05D+00,ju5)
	   uwgt5 = (0.05D+00-cumfu(ju5))/(cumfu(ju5+1)-cumfu(ju5))
	   call hunt(cumfu,nk2pts,0.1D+00,ju10)
	   uwgt10 = (0.1D+00-cumfu(ju10))/(cumfu(ju10+1)-cumfu(ju10))
	   xk5e(ii) = (one-ewgt5)*xk2pts(je5)+ewgt5*xk2pts(je5+1)
	   xk10e(ii) = (one-ewgt10)*xk2pts(je10)+ewgt10*xk2pts(je10+1)
	   xk5u(ii) = (one-uwgt5)*xk2pts(ju5)+uwgt5*xk2pts(ju5+1)
	   xk10u(ii) = (one-uwgt10)*xk2pts(ju10)+uwgt10*xk2pts(ju10+1)
	   if (nz(ii) .eq. 1) then
	      xk5eg = xk5eg+xk5e(ii)
	      xk10eg = xk10eg+xk10e(ii)
	      xk5ug = xk5ug+xk5u(ii)
	      xk10ug = xk10ug+xk10u(ii)
	   else
	      xk5eb = xk5eb+xk5e(ii)
	      xk10eb = xk10eb+xk10e(ii)
	      xk5ub = xk5ub+xk5u(ii)
	      xk10ub = xk10ub+xk10u(ii)
	   endif

         write(1,488) nz(ii),xkbaru(ii),xk2uuc,xk3uuc,xk4uuc,xk5uuc,
     &                xkbare(ii),xk2euc,xk3euc,xk4euc,xk5euc
 488     format(i4,10f25.10)

         cumf = 0.0D+00
	   cumf(1) = f0(1,1)+f0(1,2)
	   do 93 i = 2,nk2pts
	      cumf(i) = cumf(i-1)+f0(i,1)+f0(i,2)
 93      continue
         call hunt(cumf,nk2pts,0.05D+00,j5)
	   wgt5 = (0.05D+00-xk2pts(j5))/(xk2pts(j5+1)-xk2pts(j5))
	   call hunt(cumf,nk2pts,0.1D+00,j10)
	   wgt10 = (0.1D+00-xk2pts(j10))/(xk2pts(j10+1)-xk2pts(j10))
	   xk5(ii) = (one-wgt5)*xk2pts(j5)+wgt5*xk2pts(j5+1)
	   xk10(ii) = (one-wgt10)*xk2pts(j10)+wgt10*xk2pts(j10+1)
	   if (nz(ii) .eq. 1) then
	      xk5g = xk5g+xk5(ii)
	      xk10g = xk10g+xk10(ii)
	   else
	      xk5b = xk5b+xk5(ii)
	      xk10b = xk10b+xk10(ii)
	   endif

         if (nz(ii) .eq. 1) then
	      rpred(ii) = alpha*zgood*xkpred(ii)**(alpha-one)*
     &                  (hfix*(one-unempg))**(one-alpha)-delta
	      wpred(ii) = (one-alpha)*zgood*xkpred(ii)**alpha*
     &                  (hfix*(one-unempg))**(-alpha)
	      ypred(ii) = zgood*xkpred(ii)**alpha*(hfix*(one-unempg))**
     &                  (one-alpha)
	      tau = xmu*unempg/((one-unempg)*hfix)
	   else
	      rpred(ii) = alpha*zbad*xkpred(ii)**(alpha-one)*
     &                  (hfix*(one-unempb))**(one-alpha)-delta
	      wpred(ii) = (one-alpha)*zbad*xkpred(ii)**alpha*
     &                  (hfix*(one-unempb))**(-alpha)
	      ypred(ii) = zbad*xkpred(ii)**alpha*(hfix*(one-unempb))**
     &                  (one-alpha)
	      tau = xmu*unempb/((one-unempb)*hfix)
	   endif
         xpred(ii) = xkpred(ii+1)-(one-delta)*xkpred(ii)
	   cpred(ii) = ypred(ii)-xpred(ii)	   

         if (nz(ii) .eq. 1) then
	      nnnz = 2
	   else
	      nnnz = 1
	   endif
	   if (nind(ii) .eq. 1) then
	      nnni = 2
	   else
	      nnni = 1
	   endif

	   write(7,504) ii,nz(ii),xkvals(ii),
     &                stdxk(ii),skewxk(ii),xkurtk(ii),xkfive(ii),unemp,
     &                fracs(ii),abdev(ii),rent(ii),rent(ii)+delta,
     &                wage(ii),ginis(ii),gdp(ii),cons(ii),xinv(ii),
     &                pct1(ii),pct5(ii),pct10(ii),pct20(ii),pct30(ii),
     &                pct90(ii),pct95(ii),vmus(ii),xkpred(ii),
     &                xkbare(ii),xkbaru(ii),xk5e(ii),xk10e(ii),
     &                xk5u(ii),xk10u(ii),rpred(ii),wpred(ii),
     &                ypred(ii),xpred(ii),cpred(ii)
 504     format(i6,i3,35f15.9)
         write(8,556) nz(ii),nind(ii),xkind(ii),cind(ii),yind,
     &                cons(ii),gdp(ii),xkvals(ii)
 556     format(2i3,6f15.9)
         write(9,398) nnnz,nnni,xkind(ii),cind(ii),
     &                xkbaru(ii),xkbare(ii),rent(ii)+delta,wage(ii),
     &                xk5u(ii),xk10u(ii),xk5e(ii),xk10e(ii)
 398     format(2i3,10f18.10)
         write(10,399) nnnz,xkvals(ii),xkpred(ii)
 399     format(i3,2f18.10)

         if (ii .ge. 1) then
            xkmean = xkmean+xkvals(ii)/dble(real(ncapt))
            sdmean = sdmean+stdxk(ii)/dble(real(ncapt))
            skmean = skmean+skewxk(ii)/dble(real(ncapt))
            s4mean = s4mean+xkurtk(ii)/dble(real(ncapt))
	      s5mean = s5mean+xkfive(ii)/dble(real(ncapt))
	      gmean = gmean+ginis(ii)/dble(real(ncapt))
	      p1mean = p1mean+pct1(ii)/dble(real(ncapt))
	      p5mean = p5mean+pct5(ii)/dble(real(ncapt))
	      p10mean = p10mean+pct10(ii)/dble(real(ncapt))
	      p20mean = p20mean+pct20(ii)/dble(real(ncapt))
	      p30mean = p30mean+pct30(ii)/dble(real(ncapt))
	      p90mean = p90mean+pct90(ii)/dble(real(ncapt))
	      p95mean = p95mean+pct95(ii)/dble(real(ncapt))
	      contot = contot+(f0(1,1)+f0(1,2))/dble(real(ncapt))
	      if (nz(ii) .eq. 1) then
	         congd = congd+(f0(1,1)+f0(1,2))
	         mgood = mgood+1
	      else
	         conbd = conbd+(f0(1,1)+f0(1,2))
	         mbad = mbad+1
	      endif
	      xk5m = xk5m+xk5(ii)/dble(real(ncapt))
	      xk10m = xk10m+xk10(ii)/dble(real(ncapt))
		  xk5em = xk5em+xk5e(ii)/dble(real(ncapt))
	      xk10em = xk10em+xk10e(ii)/dble(real(ncapt))
	      xk5um = xk5um+xk5u(ii)/dble(real(ncapt))
	      xk10um = xk10um+xk10u(ii)/dble(real(ncapt))
	      emean1 = emean1+xkbare(ii)/dble(real(ncapt))
		  emean2 = emean2+stdxke(ii)/dble(real(ncapt))
            emean3 = emean3+skwxke(ii)/dble(real(ncapt))
	      emean4 = emean4+xkwxke(ii)/dble(real(ncapt))
	      emean5 = emean5+x5wxke(ii)/dble(real(ncapt))
	      umean1 = umean1+xkbaru(ii)/dble(real(ncapt))
	      umean2 = umean2+stdxku(ii)/dble(real(ncapt))
            umean3 = umean3+skwxku(ii)/dble(real(ncapt))
	      umean4 = umean4+xkwxku(ii)/dble(real(ncapt))
	      umean5 = umean5+x5wxku(ii)/dble(real(ncapt))
         endif

 12   continue
      close(7)
	close(8)
	close(9)
	close(10)
	close(1)

	congd = congd/dble(real(mgood))
	conbd = conbd/dble(real(mbad))
	write(6,713) contot,congd,conbd
 713  format('const: ',3f18.10)
      write(6,221) xk5em,xk10em,xk5um,xk10um
 221  format('ps: ',4f15.8)
      write(6,221) dstd(xk5e,ncapt),dstd(xk10e,ncapt),
     &             dstd(xk5u,ncapt),dstd(xk10u,ncapt)

      write(6,175)
 175  format('conditional ps')
      write(6,285) xk5eg/dble(real(mgood)),xk10eg/dble(real(mgood)),
     &             xk5ug/dble(real(mgood)),xk10ug/dble(real(mgood))
      write(6,285) xk5eb/dble(real(mbad)),xk10eb/dble(real(mbad)),
     &             xk5ub/dble(real(mbad)),xk10ub/dble(real(mbad))
      write(6,285) xk5g/dble(real(mgood)),xk10g/dble(real(mgood))
      write(6,285) xk5b/dble(real(mbad)),xk10b/dble(real(mbad))
 285  format(4f15.8)

	write(6,582)
 582  format('conditionals')
      write(6,735) emean1,emean2,emean3,emean4,emean5
	write(6,735) umean1,umean2,umean3,umean4,umean5
 735  format(5f15.8)
      write(6,701)
 701  format('conditions part 2 ')
      write(6,735) dstd(xkbare,ncapt),dstd(stdxke,ncapt),
     &             dstd(skwxke,ncapt),dstd(xkwxke,ncapt),
     &             dstd(x5wxke,ncapt)
	write(6,735) dstd(xkbaru,ncapt),dstd(stdxku,ncapt),
     &             dstd(skwxku,ncapt),dstd(xkwxku,ncapt),
     &             dstd(x5wxku,ncapt)
      write(6,124)
 124  format('unconditionals')
      write(6,735) xkmean,sdmean,skmean,s4mean,s5mean
      write(6,735) dstd(xkvals,ncapt),dstd(stdxk,ncapt),
     &             dstd(skewxk,ncapt),dstd(xkurtk,ncapt),
     &             dstd(xkfive,ncapt)
      write(6,821)
 821  format(' ')

      write(6,294) dmean(xk5,ncapt),dmean(xk10,ncapt)
	write(6,294) dstd(xk5,ncapt),dstd(xk10,ncapt)
 294  format(2f18.10)
      write(6,821)

      cgrow(1) = 0.0D+00
	do 719 i = 2,ncapt
	   cgrow(i) = cons(i)/cons(i-1)
 719  continue

c     statistics
      rhocic = correl(cind,cons,ncapt)
	rhociy = correl(cind,gdp,ncapt)
      rhocik = correl(cind,xkvals,ncapt)
	rhociyi = correl(cind,yind,ncapt)
      rhociki = correl(cind,xkind,ncapt)
	sigci = dstd(cind,ncapt)
	sigki = dstd(xkind,ncapt)
	do 682 i = 1,3
	   rhocici(i) = auto(cind,ncapt,i)
	   rhokiki(i) = auto(xkind,ncapt,i)
 682  continue
      rhocg = auto(cgrow(2:ncapt),ncapt-1,1)

      write(6,291) rhocic,rhociy,rhocik
	write(6,291) rhociyi,rhociki,sigci,sigki
	write(6,291) rhocici(1:3)
	write(6,291) rhokiki(1:3)
	write(6,291) rhocg
 291  format(4f18.10)

	stdc = dstd(cons,ncapt)
	stdy = dstd(gdp,ncapt)
	stdx = dstd(xinv,ncapt)
      
      rhoym3 = correl(gdp(4:ncapt),gdp(1:ncapt-3),ncapt-3)
      rhoym2 = correl(gdp(3:ncapt),gdp(1:ncapt-2),ncapt-2)
	rhoym1 = correl(gdp(2:ncapt),gdp(1:ncapt-1),ncapt-1)
      rhoy = correl(gdp,gdp,ncapt)
	rhoyp1 = correl(gdp(1:ncapt-1),gdp(2:ncapt),ncapt-1)
      rhoyp2 = correl(gdp(1:ncapt-2),gdp(3:ncapt),ncapt-2)
      rhoyp3 = correl(gdp(1:ncapt-3),gdp(4:ncapt),ncapt-3)

      rhocym3 = correl(gdp(4:ncapt),cons(1:ncapt-3),ncapt-3)
      rhocym2 = correl(gdp(3:ncapt),cons(1:ncapt-2),ncapt-2)
	rhocym1 = correl(gdp(2:ncapt),cons(1:ncapt-1),ncapt-1)
      rhocy = correl(gdp,cons,ncapt)
	rhocyp1 = correl(gdp(1:ncapt-1),cons(2:ncapt),ncapt-1)
      rhocyp2 = correl(gdp(1:ncapt-2),cons(3:ncapt),ncapt-2)
      rhocyp3 = correl(gdp(1:ncapt-3),cons(4:ncapt),ncapt-3)

      rhoxym3 = correl(gdp(4:ncapt),xinv(1:ncapt-3),ncapt-3)
      rhoxym2 = correl(gdp(3:ncapt),xinv(1:ncapt-2),ncapt-2)
	rhoxym1 = correl(gdp(2:ncapt),xinv(1:ncapt-1),ncapt-1)
      rhoxy = correl(gdp,xinv,ncapt)
	rhoxyp1 = correl(gdp(1:ncapt-1),xinv(2:ncapt),ncapt-1)
      rhoxyp2 = correl(gdp(1:ncapt-2),xinv(3:ncapt),ncapt-2)
      rhoxyp3 = correl(gdp(1:ncapt-3),xinv(4:ncapt),ncapt-3)

      ymean = dmean(gdp,ncapt)
	cmean = dmean(cons,ncapt)
	xmean = dmean(xinv,ncapt)

      write(6,238)
 238  format('        business cycle statistics')
      write(6,295) ymean,stdy,rhoym3,rhoym2,rhoym1,rhoy,rhoyp1,rhoyp2,
     &             rhoyp3
      write(6,295) cmean,stdc,rhocym3,rhocym2,rhocym1,rhocy,rhocyp1,
     &             rhocyp2,rhocyp3
      write(6,295) xmean,stdx,rhoxym3,rhoxym2,rhoxym1,rhoxy,rhoxyp1,
     &             rhoxyp2,rhoxyp3
 295  format(9f8.4)

	rbar = dmean(rent+delta,ncapt)
	wbar = dmean(wage,ncapt)
	stdr = dstd(rent+delta,ncapt)
	stdw = dstd(wage,ncapt)
	rhor1 = auto(rent+delta,ncapt,1)
	rhor2 = auto(rent+delta,ncapt,2)
	rhor3 = auto(rent+delta,ncapt,3)
	rhow1 = auto(wage,ncapt,1)
	rhow2 = auto(wage,ncapt,2)
	rhow3 = auto(wage,ncapt,3)
	write(6,291) rbar,stdr,wbar,stdw
      write(6,291) rhor1,rhor2,rhor3
      write(6,291) rhow1,rhow2,rhow3

	stdc = dstd(cpred,ncapt)
	stdy = dstd(ypred,ncapt)
	stdx = dstd(xpred,ncapt)
      
      rhoym3 = correl(ypred(4:ncapt),ypred(1:ncapt-3),ncapt-3)
      rhoym2 = correl(ypred(3:ncapt),ypred(1:ncapt-2),ncapt-2)
	rhoym1 = correl(ypred(2:ncapt),ypred(1:ncapt-1),ncapt-1)
      rhoy = correl(ypred,ypred,ncapt)
	rhoyp1 = correl(ypred(1:ncapt-1),ypred(2:ncapt),ncapt-1)
      rhoyp2 = correl(ypred(1:ncapt-2),ypred(3:ncapt),ncapt-2)
      rhoyp3 = correl(ypred(1:ncapt-3),ypred(4:ncapt),ncapt-3)

      rhocym3 = correl(ypred(4:ncapt),cpred(1:ncapt-3),ncapt-3)
      rhocym2 = correl(ypred(3:ncapt),cpred(1:ncapt-2),ncapt-2)
	rhocym1 = correl(ypred(2:ncapt),cpred(1:ncapt-1),ncapt-1)
      rhocy = correl(ypred,cpred,ncapt)
	rhocyp1 = correl(ypred(1:ncapt-1),cpred(2:ncapt),ncapt-1)
      rhocyp2 = correl(ypred(1:ncapt-2),cpred(3:ncapt),ncapt-2)
      rhocyp3 = correl(ypred(1:ncapt-3),cpred(4:ncapt),ncapt-3)

      rhoxym3 = correl(ypred(4:ncapt),xpred(1:ncapt-3),ncapt-3)
      rhoxym2 = correl(ypred(3:ncapt),xpred(1:ncapt-2),ncapt-2)
	rhoxym1 = correl(ypred(2:ncapt),xpred(1:ncapt-1),ncapt-1)
      rhoxy = correl(ypred,xpred,ncapt)
	rhoxyp1 = correl(ypred(1:ncapt-1),xpred(2:ncapt),ncapt-1)
      rhoxyp2 = correl(ypred(1:ncapt-2),xpred(3:ncapt),ncapt-2)
      rhoxyp3 = correl(ypred(1:ncapt-3),xpred(4:ncapt),ncapt-3)

      ymean = dmean(ypred,ncapt)
	cmean = dmean(cpred,ncapt)
	xmean = dmean(xpred,ncapt)

      write(6,237)
 237  format('        business cycle statistics (predicted)')
      write(6,295) ymean,stdy,rhoym3,rhoym2,rhoym1,rhoy,rhoyp1,rhoyp2,
     &             rhoyp3
      write(6,295) cmean,stdc,rhocym3,rhocym2,rhocym1,rhocy,rhocyp1,
     &             rhocyp2,rhocyp3
      write(6,295) xmean,stdx,rhoxym3,rhoxym2,rhoxym1,rhoxy,rhoxyp1,
     &             rhoxyp2,rhoxyp3

	rbar = dmean(rpred+delta,ncapt)
	wbar = dmean(wpred,ncapt)
	stdr = dstd(rpred+delta,ncapt)
	stdw = dstd(wpred,ncapt)
	rhor1 = auto(rpred+delta,ncapt,1)
	rhor2 = auto(rpred+delta,ncapt,2)
	rhor3 = auto(rpred+delta,ncapt,3)
	rhow1 = auto(wpred,ncapt,1)
	rhow2 = auto(wpred,ncapt,2)
	rhow3 = auto(wpred,ncapt,3)
	write(6,291) rbar,stdr,wbar,stdw
      write(6,291) rhor1,rhor2,rhor3
      write(6,291) rhow1,rhow2,rhow3

c     OLS regressions to determine implied law of motion

      npts = 0
	do 34 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdata(npts,1) = one
	      xdata(npts,2) = dlog(xkvals(i))
            ydata(npts) = dlog(xkvals(i+1))
	   endif
 34   continue
      call ols2(ydata,xdata,ncapt,a0new,rsqa,sderra,npts)

	npts = 0
	do 33 i = ndrop+1,ncapt-1
         if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdata(npts,1) = one
	      xdata(npts,2) = dlog(xkvals(i))
	      ydata(npts) = dlog(xkvals(i+1))
	   endif
 33   continue
      call ols2(ydata,xdata,ncapt,b0new,rsqb,sderrb,npts)

c     den Haan-Marcet statistic with chi-square(28) distribution

      npts = 0
	do 39 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdata2(npts,1) = one
	      xdata2(npts,2) = dlog(xkvals(i))
	      xdata2(npts,3) = xdata2(npts,2)*xdata2(npts,2)
	      xdata2(npts,4) = dlog(stdxk(i))
	      xdata2(npts,5) = xdata2(npts,4)*xdata2(npts,4)
	      xdata2(npts,6) = xdata2(npts,2)*xdata2(npts,4)
	      xdata2(npts,7) = skewxk(i)
	      xdata2(npts,8) = xdata2(npts,7)*xdata2(npts,7)
	      xdata2(npts,9) = dlog(xkurtk(i))
	      xdata2(npts,10) = xdata2(npts,9)*xdata2(npts,9)
	      xdata2(npts,11) = xdata2(npts,2)*xdata2(npts,7)
	      xdata2(npts,12) = xdata2(npts,2)*xdata2(npts,9)
	      xdata2(npts,13) = xdata2(npts,4)*xdata2(npts,7)
	      xdata2(npts,14) = xdata2(npts,4)*xdata2(npts,9)
	      xdata2(npts,15) = xdata2(npts,7)*xdata2(npts,9)
	      xdata2(npts,16) = dlog(fracs(i))
	      xdata2(npts,17) = xdata2(npts,16)*xdata2(npts,16)
	      xdata2(npts,18) = xdata2(npts,16)*xdata2(npts,2)
	      xdata2(npts,19) = xdata2(npts,16)*xdata2(npts,4)
	      xdata2(npts,20) = xdata2(npts,16)*xdata2(npts,7)
	      xdata2(npts,21) = xdata2(npts,16)*xdata2(npts,9)
	      xdata2(npts,22) = dlog(abdev(i))
	      xdata2(npts,23) = xdata2(npts,22)*xdata2(npts,22)
	      xdata2(npts,24) = xdata2(npts,22)*xdata2(npts,2)	      
	      xdata2(npts,25) = xdata2(npts,22)*xdata2(npts,4)
	      xdata2(npts,26) = xdata2(npts,22)*xdata2(npts,7)
	      xdata2(npts,27) = xdata2(npts,22)*xdata2(npts,9)
	      xdata2(npts,28) = xdata2(npts,22)*xdata2(npts,16)
         endif
 39   continue

      npts = 0
	do 98 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdata3(npts,1) = one
	      xdata3(npts,2) = dlog(xkvals(i))
	      xdata3(npts,3) = xdata3(npts,2)*xdata3(npts,2)
	      xdata3(npts,4) = dlog(stdxk(i))
	      xdata3(npts,5) = xdata3(npts,4)*xdata3(npts,4)
	      xdata3(npts,6) = xdata3(npts,2)*xdata3(npts,4)
	      xdata3(npts,7) = skewxk(i)
	      xdata3(npts,8) = xdata3(npts,7)*xdata3(npts,7)
	      xdata3(npts,9) = dlog(xkurtk(i))
	      xdata3(npts,10) = xdata3(npts,9)*xdata3(npts,9)
	      xdata3(npts,11) = xdata3(npts,2)*xdata3(npts,7)
	      xdata3(npts,12) = xdata3(npts,2)*xdata3(npts,9)
	      xdata3(npts,13) = xdata3(npts,4)*xdata3(npts,7)
	      xdata3(npts,14) = xdata3(npts,4)*xdata3(npts,9)
	      xdata3(npts,15) = xdata3(npts,7)*xdata3(npts,9)
	      xdata3(npts,16) = dlog(fracs(i))
	      xdata3(npts,17) = xdata3(npts,16)*xdata3(npts,16)
	      xdata3(npts,18) = xdata3(npts,16)*xdata3(npts,2)
	      xdata3(npts,19) = xdata3(npts,16)*xdata3(npts,4)
	      xdata3(npts,20) = xdata3(npts,16)*xdata3(npts,7)
	      xdata3(npts,21) = xdata3(npts,16)*xdata3(npts,9)
	      xdata3(npts,22) = dlog(abdev(i))
	      xdata3(npts,23) = xdata3(npts,22)*xdata3(npts,22)
	      xdata3(npts,24) = xdata3(npts,22)*xdata3(npts,2)	      
	      xdata3(npts,25) = xdata3(npts,22)*xdata3(npts,4)
	      xdata3(npts,26) = xdata3(npts,22)*xdata3(npts,7)
	      xdata3(npts,27) = xdata3(npts,22)*xdata3(npts,9)
	      xdata3(npts,28) = xdata3(npts,22)*xdata3(npts,16)
         endif
 98   continue

	nz1 = 0
	nz0 = 0
	bt = 0.0D+00
	at = 0.0D+00
	capt = dble(real(ncapt-ndrop-1))
	erfmx1 = 0.0D+00
      erfmx2 = 0.0D+00
	do 78 i = ndrop,ncapt-1
	   if (nz(i) .eq. 1) then
	      nz1 = nz1+1
	      error = dlog(xkvals(i+1))-a0new(1)-
     &              a0new(2)*dlog(xkvals(i))
	      do 77 j = 1,28
	         bt(j,1) = bt(j,1)+error*xdata2(nz1,j)/capt
	         btt(1,j) = bt(j,1)
               do 76 k = 1,28
	            at(j,k) = at(j,k)+error*error*xdata2(nz1,j)*
     &                      xdata2(nz1,k)/capt
 76            continue
 77         continue
            if (dabs(error) .ge. erfmx1) erfmx1 = dabs(error)
         else
	      nz0 = nz0+1
	      error = dlog(xkvals(i+1))-b0new(1)-
     &              b0new(2)*dlog(xkvals(i))
	      do 79 j = 1,28
	         bt(j,1) = bt(j,1)+error*xdata3(nz0,j)/capt
	         btt(1,j) = bt(j,1)
	         do 80 k = 1,28
	            at(j,k) = at(j,k)+error*error*xdata3(nz0,j)*
     &                      xdata3(nz0,k)/capt
 80            continue
 79         continue
            if (dabs(error) .ge. erfmx2) erfmx2 = dabs(error)
         endif
 78   continue
      call inv28(at,atinv,nsing)
	temp = matmul(btt,atinv)
	ttt = matmul(temp,bt)
	hansen = ttt(1,1)*capt

c     Naive robustness test

      npts = 0
	do 70 i = ndrop,ncapt-1
	   npts = npts+1
	   if (nz(i) .eq. 1) then
            error = dlog(xkvals(i+1))-a0new(1)-a0new(2)*dlog(xkvals(i))
	   else
	      error = dlog(xkvals(i+1))-b0new(1)-b0new(2)*dlog(xkvals(i))
	   endif
	   xdatan(npts,1) = dlog(abdev(i))
	   xdatan(npts,2) = dlog(stdxk(i))
	   xdatan(npts,3) = dlog(skewxk(i))
	   xdatan(npts,4) = dlog(xkurtk(i))
	   xdatan(npts,5) = dlog(ginis(i))
	   xdatan(npts,6) = dlog(fracs(i))
	   ydatan(npts) = error
 70   continue
      call ols6(ydatan,xdatan,ncapt,bhat,rsqe,sderre,npts)

      write(6,289) 
 289  format('naive test statistics')
	write(6,729) (bhat(i),i=1,6)
 729  format(3f18.10)
      write(6,729) rsqe,sderre

      npts = 0
	do 354 i = ndrop+1,ncapt-1
	   if (nz(i) .eq. 1) then
	      npts = npts+1
	      xdatas(npts,1) = one
	      xdatas(npts,2) = dlog(xkvals(i))
	      xdatas(npts,3) = dlog(stdxk(i))
	      xdatas(npts,4) = dlog(skewxk(i))
	      xdatas(npts,5) = dlog(xkurtk(i))
	      xdatas(npts,6) = dlog(ginis(i))
	      xdatas(npts,7) = dlog(fracs(i))
	      xdatas(npts,8) = dlog(abdev(i))
            ydata(npts) = dlog(stdxk(i+1))
	   endif
 354  continue
      call ols8(ydata,xdatas,ncapt,p0new,rsqp,sderrp,npts)

	npts = 0
	do 353 i = ndrop+1,ncapt-1
         if (nz(i) .eq. 2) then
	      npts = npts+1
	      xdatas(npts,1) = one
	      xdatas(npts,2) = dlog(xkvals(i))
	      xdatas(npts,3) = dlog(stdxk(i))
	      xdatas(npts,4) = dlog(skewxk(i))
	      xdatas(npts,5) = dlog(xkurtk(i))
	      xdatas(npts,6) = dlog(ginis(i))
	      xdatas(npts,7) = dlog(fracs(i))
	      xdatas(npts,8) = dlog(abdev(i))
            ydata(npts) = dlog(stdxk(i+1))
	   endif
 353  continue
      call ols8(ydata,xdatas,ncapt,q0new,rsqq,sderrq,npts)

      write(6,299) 
 299  format('naive test statistics for Sigma')
	write(6,739) (p0new(i),i=1,4)
 739  format(4f18.10)
	write(6,739) (p0new(i),i=5,8)
      write(6,729) rsqp,sderrp      
	write(6,739) (q0new(i),i=1,4)
	write(6,739) (q0new(i),i=5,8)
      write(6,729) rsqq,sderrq      

c     GMM estimation of preference parameters from aggregate consumption and return data

      vhat(1,1) = one
	vhat(1,2) = 0.0D+00
	vhat(1,3) = 0.0D+00
	vhat(2,1) = 0.0D+00
	vhat(2,2) = one
	vhat(2,3) = 0.0D+00
	vhat(1,3) = 0.0D+00
	vhat(2,3) = 0.0D+00
	vhat(3,3) = one
      call inv3(vhat,wght,nsing)

      p(1,1) = beta
	p(1,2) = gamma
	p(2,1) = 0.8D+00
	p(2,2) = gamma
	p(3,1) = beta
	p(3,2) = 4.0D+00
	do 67 i = 1,3
	   y(i) = gmm(p(i,:))
 67   continue
	call amoeba(p,y,3,2,2,1.0D-08,gmm,niter)
	bhat2 = p(1,1)
	ghat2 = p(1,2)

      vhat = 0.0D+00
      do 56 i = ndrop+1,ncapt-5
	   err1 = one-bhat2*(cons(i+1)/cons(i))**(-ghat2)*(one+rent(i+1))
	   err2 = err1*rent(i)
	   err3 = err1*cons(i)/cons(i-1)
	   errm11 = one-bhat2*(cons(i)/cons(i-1))**(-ghat2)*(one+rent(i))
	   errm21 = one-bhat2*(cons(i-1)/cons(i-2))**(-ghat2)*
     &            (one+rent(i-1))
	   errm31 = one-bhat2*(cons(i-2)/cons(i-3))**(-ghat2)*
     &            (one+rent(i-2))
	   errp11 = one-bhat2*(cons(i+2)/cons(i+1))**(-ghat2)*
     &            (one+rent(i+1))
	   errp21 = one-bhat2*(cons(i+3)/cons(i+2))**(-ghat2)*
     &            (one+rent(i+2))
	   errp31 = one-bhat2*(cons(i+4)/cons(i+3))**(-ghat2)*
     &            (one+rent(i+3))
	   errm12 = errm11*rent(i-1)
	   errm22 = errm21*rent(i-2)
	   errm32 = errm31*rent(i-3)
	   errp12 = errp11*rent(i+1)
	   errp22 = errp21*rent(i+2)
	   errp32 = errp31*rent(i+3)
	   errm13 = errm13*cons(i-1)/cons(i-2)
	   errm23 = errm23*cons(i-2)/cons(i-3)
	   errm33 = errm33*cons(i-3)/cons(i-4)
	   errp13 = errp13*cons(i+1)/cons(i)
	   errp23 = errp23*cons(i+2)/cons(i+1)
	   errp33 = errp33*cons(i+3)/cons(i+2)

         vhat(1,1) = vhat(1,1)+err1*err1+
     &               0.75D+00*(errm11*errm11+errp11*errp11)+
     &               0.50D+00*(errm21*errm21+errp21*errp21)+
     &               0.25D+00*(errm31*errm31+errp31*errp31)
	   vhat(1,2) = vhat(1,2)+err1*err2+
     &               0.75D+00*(errm11*errm12+errp11*errp12)+
     &               0.50D+00*(errm21*errm22+errp21*errp22)+
     &               0.25D+00*(errm31*errm32+errp31*errp32)
	   vhat(1,3) = vhat(1,3)+err1*err3+
     &               0.75D+00*(errm11*errm13+errp11*errp13)+
     &               0.50D+00*(errm21*errm23+errp21*errp23)+
     &               0.25D+00*(errm31*errm33+errp31*errp33)
	   vhat(2,1) = vhat(2,1)+err2*err1+
     &               0.75D+00*(errm12*errm11+errp12*errp11)+
     &               0.50D+00*(errm22*errm21+errp22*errp21)+
     &               0.25D+00*(errm32*errm31+errp32*errp31)
	   vhat(2,2) = vhat(2,2)+err2*err2+
     &               0.75D+00*(errm12*errm12+errp12*errp12)+
     &               0.50D+00*(errm22*errm22+errp22*errp22)+
     &               0.25D+00*(errm32*errm32+errp32*errp32)
	   vhat(2,3) = vhat(2,3)+err2*err3+
     &               0.75D+00*(errm12*errm13+errp12*errp13)+
     &               0.50D+00*(errm22*errm23+errp22*errp23)+
     &               0.25D+00*(errm32*errm33+errp32*errp33)
	   vhat(3,1) = vhat(3,1)+err3*err1+
     &               0.75D+00*(errm13*errm11+errp13*errp11)+
     &               0.50D+00*(errm23*errm21+errp23*errp21)+
     &               0.25D+00*(errm33*errm31+errp33*errp31)
	   vhat(3,2) = vhat(3,2)+err3*err2+
     &               0.75D+00*(errm13*errm12+errp13*errp12)+
     &               0.50D+00*(errm23*errm22+errp23*errp22)+
     &               0.25D+00*(errm33*errm32+errp33*errp32)
	   vhat(3,3) = vhat(3,3)+err3*err3+
     &               0.75D+00*(errm13*errm13+errp13*errp13)+
     &               0.50D+00*(errm23*errm23+errp23*errp23)+
     &               0.25D+00*(errm33*errm33+errp33*errp33)
 56   continue
      call inv3(vhat,wght,nsing)

      p(1,1) = bhat2
	p(1,2) = ghat2
	p(2,1) = 0.8D+00
	p(2,2) = ghat2
	p(3,1) = bhat2
	p(3,2) = 4.0D+00
	do 60 i = 1,3
	   y(i) = gmm(p(i,:))
 60   continue
	call amoeba(p,y,3,2,2,1.0D-08,gmm,niter)
	bhat2 = p(1,1)
	ghat2 = p(1,2)

      write(6,974)
 974  format('GMM estimates')
      write(6,737) beta,bhat2,gamma,ghat2
	write(6,737) y(1)
 737  format(4f18.10)

      err1mx = 0.0D+00
	err2mx = 0.0D+00
	err1sm = 0.0D+00
	err2sm = 0.0D+00
	do 190 i = 1,ncapt
	   abserr = dabs(xkvals(i)-xkpred(i))
	   relerr = dabs((xkvals(i)-xkpred(i))/xkvals(i))
	   if (abserr .gt. err1mx) err1mx = abserr
	   if (relerr .gt. err2mx) err2mx = relerr
	   err1sm = err1sm+abserr/dble(real(ncapt))
	   err2sm = err2sm+relerr/dble(real(ncapt))
 190  continue
      write(6,253) err1mx,err2mx
 253  format('max errors, random: ',2f18.10)
      write(6,252) err1sm,err2sm
 252  format('avg errors, random: ',2f18.10)

      return
      end